/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Mark Spencer <markster@digium.com>
 *
 * Licensed exclusively for use with Digium/Asterisk products.
 *
 * Copyright (C) 2008, Digium, Inc.
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 */

/*! \file
 *
 * \brief Skype For Asterisk Channel Driver
 * 
 * \ingroup channel_drivers
 */

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 2452 $")

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dlfcn.h>

#include "asterisk/localtime.h"
#include "asterisk/lock.h"
#include "asterisk/utils.h"
#include "asterisk/channel.h"
#include "asterisk/config.h"
#include "asterisk/logger.h"
#include "asterisk/module.h"
#include "asterisk/pbx.h"
#include "asterisk/options.h"
#include "asterisk/cli.h"
#include "asterisk/callerid.h"
#include "asterisk/features.h"
#include "asterisk/astobj2.h"
#include "asterisk/musiconhold.h"
#include "asterisk/devicestate.h"
#include "asterisk/paths.h"
#include "asterisk/manager.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/app.h"
#include "asterisk/astdb.h"
#include "asterisk/causes.h"

#include "version.h"
#include "skypeforasterisk.h"

#define DEFAULT_CONTEXT		"default"
#define DEFAULT_MOHINTERPRET    "default"
#define DEFAULT_MOHSUGGEST      ""

#define MAX_TIMESTAMP_SKEW	640

#define USER_BUCKETS 17

void __attribute__((weak)) ao2_iterator_destroy(struct ao2_iterator *i);

struct ao2_container *users;

enum call_direction {
	DIRECTION_INCOMING = (1 << 0),
	DIRECTION_OUTGOING = (1 << 1),
	DIRECTION_BOTH = (3 << 0),
};

enum auth_policy_type {
	AUTH_NONE = 0,
	AUTH_ACCEPT,
	AUTH_DENY,
	AUTH_BLOCK,
	AUTH_IGNORE,
};

struct auth_policy {
	enum auth_policy_type type;
	char secret[PATH_MAX];
	AST_LIST_ENTRY(auth_policy) next;
};

struct client_user_info {
	/* there are some fields of this structure that are set when it is created
	 * and will never change, thus no lock is needed to read them. they are:
	 * 	- name
	 *	- sfa_user
	 *	- sfa_socket
	 */
	struct sfa_user_info *sfa_user;
	struct ao2_container *calls;
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);
		AST_STRING_FIELD(secret);
		AST_STRING_FIELD(context);
		AST_STRING_FIELD(exten);
		AST_STRING_FIELD(accountcode);
		AST_STRING_FIELD(language);
		AST_STRING_FIELD(mohinterpret);	/*!< MOH class to use when put on hold */
		AST_STRING_FIELD(mohsuggest);	/*!< MOH class to suggest when putting a peer on hold */
		AST_STRING_FIELD(fullname);
		AST_STRING_FIELD(country);
		AST_STRING_FIELD(province);
		AST_STRING_FIELD(city);
		AST_STRING_FIELD(phone_home);
		AST_STRING_FIELD(phone_office);
		AST_STRING_FIELD(phone_mobile);
		AST_STRING_FIELD(email);
		AST_STRING_FIELD(homepage);
		AST_STRING_FIELD(about);
		AST_STRING_FIELD(avatar_filename);
		AST_STRING_FIELD(birthday);
		AST_STRING_FIELD(gender);
		AST_STRING_FIELD(timezone);
		AST_STRING_FIELD(mood_text);
		AST_STRING_FIELD(autoreply);
		AST_STRING_FIELD(webpresence);
	);
	AST_LIST_HEAD_NOLOCK(auth_policy_list, auth_policy) auth_policies;
	struct ast_codec_pref prefs;
	int sfa_socket;				/* socket used by sfa to talk to Skype engine; do not
						 * read or write this socket, the only permitted use is
						 * to call poll() to watch for events to be processed
						 */
	format_t capability;
	int amaflags;
	enum call_direction direction;
	unsigned int deletepending:1;		/* during a reload, all users are marked for deletion until found
						 * in the configuration file
						 */
	unsigned int debug:1;			/* debug messages should be generated for actions on this user */
	unsigned int vedebug:1;			/* voiceengine debug messages should be generated for actions on this user */
	unsigned int buddy_presence:1;
	/* TODO: finish broken state support */
	unsigned int broken:1;			/* a poll() on this user's sfa_socket failed previously; only set by do_monitor() */
	unsigned int autoadd_hint_contacts:1;
	unsigned int autoadd_buddy_contacts:1;
	unsigned int fresh:1;
	unsigned int login_queued:1;
	unsigned int property_update_queued:1;
};

struct client_call_info {
	/* there are some fields of this structure that are set when it is created
	 * and will never change, thus no lock is needed to read them. they are:
	 *	- sfa_call
	 *	- user
	 */
	struct sfa_call_info *sfa_call;
	struct ast_channel *owner;
	struct client_user_info *user;
	struct ast_rtp_instance *rtp;
	char *cid_num;
	char *cid_name;
	int control_socket;
	int chanalert;
	unsigned int sent_ringing:1;
	unsigned int got_codec:1;
	unsigned int delay_answer_until_codec:1;
	unsigned int call_members_updated:1;
	unsigned int client_call_set:1;
	char ident[256];
	struct ast_frame f;
	enum call_direction direction;
};

struct global_config_options {
	char engine_directory[PATH_MAX];
	char data_directory[PATH_MAX];
	char defaultuser[PATH_MAX];
	char bind_address[16];
	char rtp_address[16];
	char bind_port[6];
	char https_proxy[PATH_MAX];
	char https_proxy_user[PATH_MAX];
	char https_proxy_password[PATH_MAX];
	char socks5_proxy[PATH_MAX];
	char socks5_proxy_user[PATH_MAX];
	char socks5_proxy_password[PATH_MAX];
	unsigned int disable_tcpauto:1;
	unsigned int disable_udp:1;
	unsigned int debug:1;
	struct sockaddr_in bind_sockaddr_in;
	struct sockaddr_in rtp_sockaddr_in;
};

static struct global_config_options global_defaults = {
	.engine_directory = "/tmp",
	.bind_address = "0.0.0.0",
	.rtp_address = "127.0.0.1",
	.bind_port = "0",
};

struct config_option {
	const char *name;
	const char *default_value;
	unsigned int parsed:1;
};

enum {
	FD_RTP = 0,
	FD_VE,
	FD_CHAN,
};

#define INIT_CONFIG_OPTION(__options, __name, __default) \
	do { \
		struct config_option *__o; \
		if (!(__o = ao2_alloc(sizeof(*__o), config_option_destructor))) { \
			break; \
		} \
		__o->name = (__name); \
		__o->default_value = (__default);\
		ao2_link((__options), __o); \
		ao2_ref(__o, -1); \
	} while (0);

static int skype_transfer_exec(struct ast_channel *chan, const char *data);

static int option_hash_cb(const void *obj, const int flags)
{
	const struct config_option *opt = obj;
	return ast_str_case_hash(opt->name);
}

static int option_cmp_cb(void *obj, void *arg, int flags)
{
	struct config_option *opt1 = obj, *opt2 = arg;
	return !strcasecmp(opt1->name, opt2->name) ? CMP_MATCH | CMP_STOP : 0;
}

static void config_option_destructor(void *obj)
{
	if (option_debug > 3) {
		ast_log(LOG_DEBUG, "Destroing %p\n", obj);
	}
}

static struct config_option *find_option(struct ao2_container *con, const char *name)
{
	struct config_option search = {
		.name = name,
	};
	return ao2_find(con, &search, OBJ_POINTER);
}

static struct global_config_options global_config;
static int broken;

static void call_datastore_free(void *data);
static void *call_datastore_duplicate(void *data);

static struct ast_datastore_info skype_call_datastore = {
	.type = "SKYPE_CALL",
	.destroy = call_datastore_free,
	.duplicate = call_datastore_duplicate,
};

static void call_datastore_free(void *data)
{
	struct varshead *varshead = data;
	struct ast_var_t *var;

	while ((var = AST_LIST_REMOVE_HEAD(varshead, entries))) {
		ast_var_delete(var);
	}
	ast_free(varshead);
}

static void *call_datastore_duplicate(void *data)
{
	struct varshead *copy, *orig = data;
	struct ast_var_t *copy_var, *var;

	if (!(copy = ast_calloc(1, sizeof(copy)))) {
		return NULL;
	}

	AST_LIST_TRAVERSE(orig, var, entries) {
		copy_var = ast_var_assign(ast_var_name(var), ast_var_value(var));
		AST_LIST_INSERT_HEAD(copy, copy_var, entries);
	}
	return copy;
}

static struct ast_codec_pref default_prefs;

#define GLOBAL_DEBUG (global_config.debug)
#define USER_DEBUG(u) (u->debug)
#define USER_VEDEBUG(u) (u->vedebug)

static pthread_t monitor_thread = AST_PTHREADT_NULL;
static unsigned int monitor_shutdown;
static unsigned int monitor_reload;
static int monitor_alert[2];

static unsigned int dynamic_codec_change;

static struct ast_channel *new_ast_channel(struct client_call_info *i, enum ast_channel_state state, const char *linkedid);
static struct ast_channel *request(const char *type, format_t format, const struct ast_channel *requestor, void *data, int *cause);
static int digit_begin(struct ast_channel *ast, char digit);
static int digit_end(struct ast_channel *ast, char digit, unsigned int duration);
static int call(struct ast_channel *ast, char *dest, int timeout);
static int hangup(struct ast_channel *ast);
static int answer(struct ast_channel *ast);
static struct ast_frame *frame_read(struct ast_channel *ast);
static int frame_write(struct ast_channel *ast, struct ast_frame *frame);
static int indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen);
static int fixup(struct ast_channel *oldchan, struct ast_channel *newchan);
static int channel_read(struct ast_channel *chan, const char *funcname, char *preparse, char *buf, size_t buflen);
static int transfer(struct ast_channel *chan, const char *destination);

static void destroy_client_call(void *obj);
static void destroy_client_user(void *obj);
static void unlink_client_call(struct client_call_info *pvt);
static void *do_monitor(void *data);

static enum ast_rtp_glue_result rtp_get_info(struct ast_channel *chan, struct ast_rtp_instance **instance);
static int rtp_set_peer(struct ast_channel *chan, struct ast_rtp_instance *peer, struct ast_rtp_instance *vpeer, struct ast_rtp_instance *tpeer, format_t codecs, int nat_active);
static format_t rtp_get_codec(struct ast_channel *chan);

static struct ast_channel_tech chan_tech = {
	.type = "Skype",
	.description = "Skype For Asterisk Channel Driver",
	.capabilities = AST_FORMAT_SLINEAR,
	.requester =  request,
	.send_digit_begin = digit_begin,
	.send_digit_end = digit_end,
	.call = call,
	.hangup = hangup,
	.answer = answer,
	.read = frame_read,
	.write = frame_write,
	.indicate = indicate,
	.fixup = fixup,
	.func_channel_read = channel_read,
	.bridge = ast_rtp_instance_bridge,
	.transfer = transfer,
};

static struct ast_rtp_glue skype_rtp_glue = {
	.type = "Skype",
	.get_rtp_info = rtp_get_info,
	.update_peer = rtp_set_peer,
	.get_codec = rtp_get_codec,
};

/* Forward declarations */
static struct client_user_info *find_active_user(const char *name);

static struct ao2_container *unref_users(struct ao2_container *users)
{
	ao2_ref(users, -1);
	return NULL;
}

static int user_hash_cb(const void *obj, const int flags)
{
	const struct client_user_info *user = obj;

	return ast_str_hash(user->name);
}

static int user_cmp_cb(void *obj, void *arg, int flags)
{
	const struct client_user_info *user1 = obj;
	const struct client_user_info *user2 = arg;

	/* all calls to this function will pass a pointer to a user
	 * object as the arg to search for, so we can optimize
	 * the matching process
	 */

	return !strcmp(user1->name, user2->name) ? CMP_MATCH | CMP_STOP : 0;
}

static int call_hash_cb(const void *obj, const int flags)
{
	const struct client_call_info *call = obj;

	return ast_str_hash(call->user->name);
}

static enum ast_rtp_glue_result rtp_get_info(struct ast_channel *chan, struct ast_rtp_instance **instance)
{
	struct client_call_info *p;

	if (!(p = (struct client_call_info *) chan->tech_pvt)) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	ao2_ref(p->rtp, +1);
	*instance = p->rtp;

	return AST_RTP_GLUE_RESULT_FORBID;
}

static int rtp_set_peer(struct ast_channel *chan, struct ast_rtp_instance *peer, struct ast_rtp_instance *vpeer, struct ast_rtp_instance *tpeer, format_t codecs, int nat_active)
{
	/* direct media path is not possible */
	return 0;
}

static format_t rtp_get_codec(struct ast_channel *chan)
{
	return chan->nativeformats;
}

static int indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen)
{
	struct client_call_info *p = (struct client_call_info *) ast->tech_pvt;

	switch (condition) {
	case AST_CONTROL_RINGING:
		if ((ast->_state == AST_STATE_RING) && !p->sent_ringing) {
			/* Treat like an answer */
			sfa_call_ringback(p->sfa_call);
			p->sent_ringing = 1;
			return 0;
		}
		return -1;
	case AST_CONTROL_HOLD:
	{
		char *tmp;

		/* hold the user lock for the shortest time possible */
		ao2_lock(p->user);
		tmp = ast_strdupa(p->user->mohinterpret);
		ao2_unlock(p->user);
		if (strcasecmp(tmp, "passthrough")) {
			ast_rtp_instance_update_source(p->rtp);
			ast_moh_start(ast, data, tmp);
		} else {
			sfa_call_hold(p->sfa_call);
		}
		return 0;
	}
	case AST_CONTROL_UNHOLD:
	{
		char *tmp;

		/* hold the user lock for the shortest time possible */
		ao2_lock(p->user);
		tmp = ast_strdupa(p->user->mohinterpret);
		ao2_unlock(p->user);
		if (strcasecmp(tmp, "passthrough")) {
			ast_rtp_instance_update_source(p->rtp);
			ast_moh_stop(ast);
		} else {
			sfa_call_unhold(p->sfa_call);
		}
		return 0;
	}
	case AST_CONTROL_SRCUPDATE:
		ast_rtp_instance_update_source(p->rtp);
		return 0;
	default:
		return -1;
	}
}

static int transfer(struct ast_channel *chan, const char *destination)
{
	return skype_transfer_exec(chan, (void *) destination);
}

static int fixup(struct ast_channel *oldchan, struct ast_channel *newchan)
{
	struct client_call_info *p = (struct client_call_info *) newchan->tech_pvt;

	p->owner = newchan;

	return 0;
}

static int digit_begin(struct ast_channel *ast, char digit)
{
	struct client_call_info *p = (struct client_call_info *) ast->tech_pvt;

	ast_rtp_instance_dtmf_begin(p->rtp, digit);

	return 0;
}

static int digit_end(struct ast_channel *ast, char digit, unsigned int duration)
{
	struct client_call_info *p = (struct client_call_info *) ast->tech_pvt;

	ast_rtp_instance_dtmf_end(p->rtp, digit);

	return 0;
}

static int channel_read(struct ast_channel *chan, const char *funcname, char *preparse, char *buf, size_t buflen)
{
	enum sfa_property prop;
	struct client_call_info *client_call = chan->tech_pvt;
	char *property;

	/* Sanity check */
	if (chan->tech != &chan_tech) {
		ast_log(LOG_ERROR, "Cannot call %s on a non-%s channel\n", funcname, chan_tech.type);
		return 0;
	}

	if (!strncasecmp(preparse, "skype_", 6)) {
		property = preparse + 6;
	} else {
		property = preparse;
	}

	if (!(prop = sfa_str2property(property))) {
		ast_log(LOG_WARNING, "Unrecognized argument '%s' to %s\n", preparse, funcname);
		return 0;
	}

	memset(buf, 0, buflen);

	return sfa_call_get_property(client_call->sfa_call, prop, buf, buflen);
}

static void set_client_call(struct client_call_info *pvt) {
	if (pvt->client_call_set) {
		return;
	}
	pvt->client_call_set = 1;
	ao2_ref(pvt, +1);
	sfa_call_set_client_call(pvt->sfa_call, pvt);
}

static void unset_client_call(struct client_call_info *pvt) {
	sfa_call_set_client_call(pvt->sfa_call, NULL);
	ao2_ref(pvt, -1);
}

/* must be able to return an error, and destroy the call if needed */
static void update_call_members(struct client_call_info *pvt)
{
	char buf[80] = {0,};

	if (pvt->call_members_updated) {
		ast_log(LOG_WARNING, "Whoa, call members already updated!\n");
		return;
	}

	set_client_call(pvt);
	if (!sfa_call_get_property(pvt->sfa_call, SFA_SKYPENAME, buf, sizeof(buf))) {
		pvt->cid_num = ast_strdup(buf);
	}
	if (ast_strlen_zero(buf) &&
		!sfa_call_get_property(pvt->sfa_call, SFA_PSTNNUMBER, buf, sizeof(buf))) {
		pvt->cid_num = ast_strdup(buf);
	}
	if (!sfa_call_get_property(pvt->sfa_call, SFA_FULLNAME, buf, sizeof(buf))) {
		pvt->cid_name = ast_strdup(buf);
	}
	pvt->call_members_updated = 1;
}

static enum sfa_property sfa_call_property_from_name(const char *name)
{
	if (!strcasecmp(name, "topic")) {
		return SFA_CALL_TOPIC;
	} else if (!strcasecmp(name, "token")) {
		return SFA_CALL_TOKEN;
	} else {
		return 0;
	}
}

static int call(struct ast_channel *ast, char *dest, int timeout)
{
	struct client_call_info *p = (struct client_call_info *) ast->tech_pvt;
 	struct ast_datastore *datastore;

	if (!p->user || !p->user->sfa_user || !sfa_user_loggedin(p->user->sfa_user)) {
		ast_log(LOG_WARNING, "User '%s' is not ready for calls yet!\n", p->user->name);
		return -1;
	}
  
	/* Check for properties that have been set via SKYPE_CALL_PROPERTY */
	if ((datastore = ast_channel_datastore_find(ast, &skype_call_datastore, "skype_call_property"))) {
		struct varshead *varshead = datastore->data;
		struct ast_var_t *v;

		if (!varshead) {
			return -1;
		}

		AST_LIST_TRAVERSE(varshead, v, entries) {
			enum sfa_property prop;
			if (!(prop = sfa_call_property_from_name(ast_var_name(v)))) {
				ast_log(LOG_WARNING, "Setting property '%s' is not supported\n", ast_var_name(v));
				continue;
			}
			sfa_call_set_property(p->sfa_call, prop, ast_var_value(v));
		}
	}

	if (!sfa_call_ring(p->sfa_call, p->ident, &p->control_socket, &p->chanalert)) {
		update_call_members(p);
		ast->fds[FD_VE] = p->control_socket;
		ast->fds[FD_CHAN] = p->chanalert;
		return 0;
	} else {
		ast_log(LOG_WARNING, "Unable to ring user '%s'\n", p->ident);
		return -1;
	}
}

static int hangup(struct ast_channel *ast)
{
	struct client_call_info *p = (struct client_call_info *) ast->tech_pvt;
	
	unlink_client_call(p);
	sfa_call_hangup(p->sfa_call);
	p->sfa_call = NULL;
	p->owner = NULL;
	ast->tech_pvt = NULL;
	ao2_ref(p, -1);

	return 0;
}

static int answer(struct ast_channel *ast)
{
	struct client_call_info *p = (struct client_call_info *) ast->tech_pvt;
	
	sfa_call_answer(p->sfa_call);

	return 0;
}

static struct ast_frame *frame_read(struct ast_channel *ast)
{
	struct client_call_info *p = (struct client_call_info *) ast->tech_pvt; 
	static struct ast_frame f = { AST_FRAME_NULL }; 
	int res;

	switch (ast->fdno) {
	case FD_RTP:
		/* the RTP socket has data available */
		return ast_rtp_instance_read(p->rtp, 0);
	case FD_VE:
		/* the control socket has data available */
		res = sfa_call_process(p->sfa_call);
		return &f;
	case FD_CHAN:
		res = sfa_channel_process(p->sfa_call);
		return &f;
	default:
		return &f;
	}
}

static int frame_write(struct ast_channel *ast, struct ast_frame *f)
{
	struct client_call_info *p = (struct client_call_info *) ast->tech_pvt;

	switch (f->frametype) {
	case AST_FRAME_VOICE:
		return ast_rtp_instance_write(p->rtp, f);
	default:
		return 0;
	}
}

static struct ast_channel *new_ast_channel(struct client_call_info *i, enum ast_channel_state state, const char *linkedid)
{
	struct ast_channel *tmp;
	int fmt;

	if (!(tmp = ast_channel_alloc(1, state, i->cid_num, i->cid_name,
				      i->user->accountcode, i->user->exten,
				      i->user->context, linkedid, i->user->amaflags,
				      "Skype/%s-%08x", i->user->name, (int)(long) i))) {
		return NULL;
	}

	tmp->nativeformats = ast_codec_choose(&i->user->prefs, i->user->capability, 1);
	fmt = ast_best_codec(tmp->nativeformats);
	tmp->writeformat = fmt;
	tmp->rawwriteformat = fmt;
	tmp->readformat = fmt;
	tmp->rawreadformat = fmt;
	
	tmp->tech_pvt = i;
	ao2_ref(i, +1);
	tmp->tech = &chan_tech;

	tmp->fds[FD_RTP] = ast_rtp_instance_fd(i->rtp, 0);
	if (i->control_socket > -1) {
		tmp->fds[FD_VE] = i->control_socket;
	}

	i->owner = tmp;

	return tmp;
}

static struct client_call_info *new_client_call(struct client_user_info *user, struct sfa_call_info *sfa_call, int update, enum call_direction direction)
{
	struct client_call_info *pvt;
	struct ast_sockaddr bindaddr_tmp;

	if (!(pvt = ao2_alloc(sizeof(*pvt), destroy_client_call))) {
		return NULL;
	}

	ast_sockaddr_from_sin(&bindaddr_tmp, &global_config.rtp_sockaddr_in);
	if (!(pvt->rtp = ast_rtp_instance_new("asterisk", NULL, &bindaddr_tmp, NULL))) {
		ast_log(LOG_ERROR, "Unable to allocate RTP structure for call.\n");
		ao2_ref(pvt, -1);
		return NULL;
	}

	ast_rtp_codecs_payloads_clear(ast_rtp_instance_get_codecs(pvt->rtp), pvt->rtp);
	ast_rtp_instance_dtmf_mode_set(pvt->rtp, AST_RTP_DTMF_MODE_RFC2833);
	ast_rtp_codecs_packetization_set(ast_rtp_instance_get_codecs(pvt->rtp), pvt->rtp, &user->prefs);

	/* configure for G.711 mu-law */
	ast_rtp_codecs_payloads_set_m_type(ast_rtp_instance_get_codecs(pvt->rtp), pvt->rtp, 0);
	ast_rtp_codecs_payloads_set_rtpmap_type(ast_rtp_instance_get_codecs(pvt->rtp), pvt->rtp, 0, "audio", "PCMU", 0);

	/* configure for G.711 A-law */
	ast_rtp_codecs_payloads_set_m_type(ast_rtp_instance_get_codecs(pvt->rtp), pvt->rtp, 8);
	ast_rtp_codecs_payloads_set_rtpmap_type(ast_rtp_instance_get_codecs(pvt->rtp), pvt->rtp, 8, "audio", "PCMA", 0);

	/* configure for G.729A */
	ast_rtp_codecs_payloads_set_m_type(ast_rtp_instance_get_codecs(pvt->rtp), pvt->rtp, 18);
	ast_rtp_codecs_payloads_set_rtpmap_type(ast_rtp_instance_get_codecs(pvt->rtp), pvt->rtp, 18, "audio", "G729A", 0);

	pvt->sfa_call = sfa_call;
	pvt->user = user;
	pvt->direction = direction;
	ao2_ref(user, +1);
	pvt->control_socket = -1;

	if (update) {
		update_call_members(pvt);
	} else {
		set_client_call(pvt);
	}

	ao2_link(user->calls, pvt);

	return pvt;	
}

static enum sfa_auth_result authrequest(struct client_user_info *user, const char *peer, const char *reason)
{
	struct auth_policy *pol;

	/* it is safe to access the user object freely here, because the only way this
	 * function can be called is by do_monitor() processing callbacks for the user,
	 * and do_monitor() also handles reloads, which are the only way the user
	 * object can change (except for reference count)
	 */

	if (!user->sfa_user) {
		return SFA_AUTH_IGNORED;
	}

	if (AST_LIST_EMPTY(&user->auth_policies)) {
		sfa_user_get_buddy_state(user->sfa_user, peer, user->autoadd_buddy_contacts);
		return SFA_AUTH_ACCEPTED;
	}
	
	AST_LIST_TRAVERSE(&user->auth_policies, pol, next) {
		switch (pol->type) {
		case AUTH_ACCEPT:
			if (ast_strlen_zero(pol->secret) ||
			    (!strcasecmp(pol->secret, reason))) {
				sfa_user_get_buddy_state(user->sfa_user, peer, user->autoadd_buddy_contacts);
				return SFA_AUTH_ACCEPTED;
			}
			break;
		case AUTH_DENY:
			return SFA_AUTH_DENIED;
		case AUTH_IGNORE:
			return SFA_AUTH_IGNORED;
		case AUTH_BLOCK:
			return SFA_AUTH_BLOCKED;
		case AUTH_NONE:
			/* wtf? this should not happen */
			continue;
		}
	}

	return SFA_AUTH_IGNORED;
}

struct chat_subscriber {
	const char *to;
	const char *from;
	char *message;
	ast_mutex_t lock;
	ast_cond_t cv;
};

static struct ao2_container *chat_subscribers;

static int chat_subscriber_hash_fn(const void *obj, const int flags)
{
	const struct chat_subscriber *mw = obj;

	return ast_str_case_hash(mw->to);
}

static int chat_subscriber_cmp_fn(void *obj, void *arg, int flags)
{
	const struct chat_subscriber *one = obj, *two = arg;

	return !strcasecmp(one->to, two->to) && !strcasecmp(one->from, two->from) ? CMP_MATCH | CMP_STOP : 0;
}

static void chat_subscriber_destructor(void *obj)
{
	struct chat_subscriber *mw = obj;

	ast_free(mw->message);
	ast_cond_destroy(&mw->cv);
	ast_mutex_destroy(&mw->lock);
}

static int chat_subscriber_match_cb(void *obj, void *arg, int flags)
{
	struct chat_subscriber *mw = obj, *in = arg;

	if (strcasecmp(mw->to, in->to) || strcasecmp(mw->from, in->from)) {
		return 0;
	}

	mw->message = ast_strdup(in->message);
	ast_mutex_lock(&mw->lock);
	ast_cond_signal(&mw->cv);
	ast_mutex_unlock(&mw->lock);

	return CMP_MATCH;
}

static void new_chat_message(const char *to, const char *from, const char *text)
{
	char *base64;
	int maxlen, srclen;
	struct client_user_info *user;
	struct chat_subscriber search = {
		.from = from,
		.to = to,
		.message = ast_strdupa(text),
	};

	/* For each matching subscriber to this message, call callback function to notify */
	ao2_callback(chat_subscribers, OBJ_NODATA | OBJ_MULTIPLE, chat_subscriber_match_cb, &search);

	srclen = strlen(text);
	maxlen = ((srclen + 2) * 4 / 3) + 1;
	base64 = alloca(maxlen);
	ast_base64encode(base64, (const unsigned char *) text, srclen, maxlen);

	manager_event(EVENT_FLAG_SYSTEM, "SkypeChatMessage", "To: %s\r\nFrom: %s\r\nMessage: %s\r\n", to, from, base64);

	if (!(user = find_active_user(to))) {
		/* No log message here, becuase we will also get the event for our
		 * own outgoing messages to users who aren't on our system */
		return;
	}

	if (!ast_strlen_zero(user->autoreply)) {
		struct client_user_info *fromuser;

		if ((fromuser = find_active_user(from)) && !ast_strlen_zero(fromuser->autoreply)) {
			ast_log(LOG_WARNING, "Refusing to send an autoreply to %s who also has autoreply enabled\n", from);
			ao2_ref(fromuser, -1);
			return;
		}
		sfa_send_chat_message(user->sfa_user, from, user->autoreply);
	}
	ao2_ref(user, -1);
}

static void queue_event(struct client_call_info *pvt, int frametype, int subclass, const char *reason)
{
	struct ast_frame f = { 0, };

	if (!pvt->owner) {
		return;
	}

	if (GLOBAL_DEBUG) {
		ast_log(LOG_DEBUG, "got control event type %d with subclass %d.\n", frametype, subclass);
	}

	switch (frametype) {
	case AST_FRAME_CONTROL:
		switch (subclass) {
		case AST_CONTROL_HANGUP:
			if (!strcasecmp(reason, "Refused")) {
				pvt->owner->hangupcause = AST_CAUSE_CALL_REJECTED;
			} else if (!strcasecmp(reason, "Cancelled")) {
				pvt->owner->hangupcause = AST_CAUSE_NORMAL;
			} else if (!strcasecmp(reason, "Missed")) {
				pvt->owner->hangupcause = AST_CAUSE_NOANSWER;
			} else if (!strcasecmp(reason, "Busy")) {
				pvt->owner->hangupcause = AST_CAUSE_BUSY;
			} else if (!strcasecmp(reason, "Dropped")) {
				pvt->owner->hangupcause = AST_CAUSE_FAILURE;
			} else if (!strcasecmp(reason, "Finished")) {
				pvt->owner->hangupcause = AST_CAUSE_NORMAL;
			} else if (!strncasecmp(reason, "Voicemail", 9)) {
				pvt->owner->hangupcause = AST_CAUSE_NOANSWER;
			} else if (!strcasecmp(reason, "No Common Codec")) {
				pvt->owner->hangupcause = AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
			} else if (!strcasecmp(reason, "Recipient Not Found") || !strcasecmp(reason, "Recipient Not Online")) {
				pvt->owner->hangupcause = AST_CAUSE_UNREGISTERED;
			} else {
				pvt->owner->hangupcause = AST_CAUSE_CONGESTION;
			}
			ast_queue_hangup(pvt->owner);
			break;
		case AST_CONTROL_ANSWER:
			/* Skype sends ANSWER even on incoming calls, which Asterisk
			   doesn't want.
			*/
			if (pvt->direction == DIRECTION_INCOMING) {
				break;
			}
			/* if we haven't received indication of the codec selected for the call
			 * yet, we don't want to forward the ANSWER into Asterisk, because that
			 * could result in unnecessary codec translation. mark the channel as
			 * 'needing ANSWER once the codec has been selected' instead.
			 */
			if (!pvt->got_codec) {
				pvt->delay_answer_until_codec = 1;
				break;
			}
			if (pvt->owner->_state != AST_STATE_UP) {
				ast_queue_control(pvt->owner, AST_CONTROL_ANSWER);
			}
			break;
		default:
			ast_queue_control(pvt->owner, subclass);
			break;
		}
		break;
	default:
		f.frametype = frametype;
		f.subclass.integer = subclass;
		ast_queue_frame(pvt->owner, &f);
	}
}

static int new_call(struct client_user_info *user, struct sfa_call_info *sfa_call, int control_socket, int chanalert)
{
	struct client_call_info *client_call;
	struct ast_channel *chan;
	
	/* it is safe to access the user object freely here, because the only way this
	 * function can be called is by do_monitor() processing callbacks for the user,
	 * and do_monitor() also handles reloads, which are the only way the user
	 * object can change (except for reference count)
	 */

	if (!(user->direction & DIRECTION_INCOMING)) {
		sfa_call_hangup(sfa_call);
		return -1;
	}

	if (!(client_call = new_client_call(user, sfa_call, 1, DIRECTION_INCOMING))) {
		ast_log(LOG_WARNING, "Unable to create new private structure!\n");
		sfa_call_hangup(sfa_call);
		return -1;
	}

	client_call->control_socket = control_socket;

	if (GLOBAL_DEBUG || USER_DEBUG(user)) {
		ast_log(LOG_DEBUG, "Incoming call for Skype user %s from %s (%s)\n",
			user->name,
			(client_call->cid_num ? client_call->cid_num : "<unknown number>"), 
			(client_call->cid_name ? client_call->cid_name : "<unknown name>"));
	}

	if (!(chan = new_ast_channel(client_call, AST_STATE_RING, NULL))) {
		ast_log(LOG_WARNING, "Unable to create channel\n");
		unlink_client_call(client_call);
		sfa_call_hangup(sfa_call);
		client_call->owner = NULL;	
		ao2_ref(client_call, -1);
		return -1;
	}

	chan->fds[FD_CHAN] = client_call->chanalert = chanalert;

	if (ast_pbx_start(client_call->owner)) {
		ast_log(LOG_WARNING, "Unable to start PBX on channel '%s'\n", client_call->owner->name);
		client_call->owner->hangupcause = AST_CAUSE_SWITCH_CONGESTION;
		ast_hangup(client_call->owner);
	}

	ao2_ref(client_call, -1);
	return 0;
}

static struct client_user_info *find_user(const char *name)
{
	struct client_user_info search = {
		.name = name,
	};
	
	return ao2_find(users, &search, OBJ_POINTER);
}

static struct client_user_info *find_active_user(const char *name)
{
	struct client_user_info *user = find_user(name);

	return user && user->sfa_user ? user : NULL;
}

static struct ast_channel *request(const char *type, format_t format, const struct ast_channel *requestor, void *vdata, int *cause)
{
	char *tmp;
	char *tok1, *tok2, *ident, *fromident;
	struct client_user_info *user;
	struct client_call_info *pvt;
	struct sfa_call_info *sfa_call;
	struct ast_channel *chan;

	if (!vdata)
		return NULL;

	tmp = ast_strdupa(vdata);
	tok1 = strsep(&tmp, "@");
	tok2 = tmp;

	if (!tok2) {
		fromident = NULL;
		ident = tok1;
	} else {
		fromident = tok1;
		ident = tok2;
	}

	if (!fromident || ast_strlen_zero(fromident)) {
		if (ast_strlen_zero(global_config.defaultuser)) {
			ast_log(LOG_ERROR, "No default user has been specified, so dialing over Skype must include a 'from' username\n");
			return NULL;
		}
		fromident = global_config.defaultuser;
	}

	if (!(user = find_active_user(fromident))) {
		ast_log(LOG_WARNING, "No active Skype user '%s'\n", fromident);
		return NULL;
	}

	if (!(user->direction & DIRECTION_OUTGOING)) {
		ast_log(LOG_WARNING, "Cannot place outgoing call for user '%s', outgoing calls disabled.\n", user->name);
		ao2_ref(user, -1);
		return NULL;
	}

	if (!(sfa_call = sfa_call_new(user->sfa_user))) {
		ast_log(LOG_WARNING, "Unable to create Skype call\n");
		ao2_ref(user, -1);
		return NULL;
	}

	ao2_lock(user);
	if (!(pvt = new_client_call(user, sfa_call, 0, DIRECTION_OUTGOING))) {
		sfa_call_hangup(sfa_call);
		ao2_unlock(user);
		ao2_ref(user, -1);
		return NULL;
	}

	if (!(chan = new_ast_channel(pvt, AST_STATE_DOWN, requestor ? requestor->linkedid : NULL))) {
		ast_log(LOG_WARNING, "Unable to create channel\n");
		ao2_ref(pvt, -1);
		ao2_unlock(user);
		ao2_ref(user, -1);
		return NULL;
	}
	ao2_unlock(user);

	ast_copy_string(pvt->ident, ident, sizeof(pvt->ident));
	ao2_ref(pvt, -1);
	ao2_ref(user, -1);

	return chan;
}

static void send_monitor_alert(void)
{
	char trigger = 'X';

	if (write(monitor_alert[1], &trigger, 1) != 1) {
		/* we can't do anything here if the write fails */
	}
}

static void destroy_client_user(void *obj)
{
	struct client_user_info *user = obj;
	struct auth_policy *pol;

	if (user->calls) {
		ao2_ref(user->calls, -1);
	}
	if (user->sfa_user) {
		sfa_user_destroy(user->sfa_user);
	}
	while ((pol = AST_LIST_REMOVE_HEAD(&user->auth_policies, next))) {
		ast_free(pol);
	}
	ast_string_field_free_memory(user);
}

static int add_user_auth_policy(struct client_user_info *user, const char *data)
{
	struct auth_policy *pol;
	char *secret = ast_strdupa(data);
	char *policy = strsep(&secret, ":");
	enum auth_policy_type type = AUTH_NONE;

	if (!strcasecmp(policy, "deny") && !secret) {
		type = AUTH_DENY;
	} else if (!strcasecmp(policy, "block") && !secret) {
		type = AUTH_BLOCK;
	} else if (!strcasecmp(policy, "ignore") && !secret) {
		type = AUTH_IGNORE;
	} else if (!strcasecmp(policy, "accept")) {
		type = AUTH_ACCEPT;
	} else {
		return -1;
	}

	if (!(pol = ast_calloc(1, sizeof(*pol)))) {
		return -1;
	}

	pol->type = type;
	if (secret) {
		ast_copy_string(pol->secret, secret, sizeof(pol->secret));
	}

	AST_LIST_INSERT_TAIL(&user->auth_policies, pol, next);

	return 0;
}

static void set_buddy_autoadd_flags(struct client_user_info *user, const char *value)
{
	char buf[80];
	char *word, *next=buf;

	user->autoadd_hint_contacts = 0;
	user->autoadd_buddy_contacts = 0;
	
	if (ast_strlen_zero(value) || ast_false(value)) {
		return;
	}

	if (ast_true(value)) {
		user->autoadd_hint_contacts = 1;
		return;
	}

	ast_copy_string(buf, value, sizeof(buf));
	while ((word = strsep(&next, ","))) {
		if (!strcasecmp(word, "buddies")) {
			user->autoadd_buddy_contacts = 1;
		} else if (!strcasecmp(word, "hints")) {
			user->autoadd_hint_contacts = 1;
		}
	}
}

static void connect_user(struct client_user_info *user)
{
	char codec_buf[128] = { 0, };
	char *codec_str = codec_buf;
	size_t codec_buf_size = sizeof(codec_buf);
	unsigned int x;

	if (!user->sfa_user) {
		ast_log(LOG_WARNING, "User '%s' failed to load, not connecting\n", user->name);
		return;
	}

	user->sfa_socket = sfa_user_get_socket(user->sfa_user);
	sfa_user_set_debug(user->sfa_user, USER_DEBUG(user));
	sfa_user_set_vedebug(user->sfa_user, USER_VEDEBUG(user));
	sfa_user_login(user->sfa_user, user->secret);

	for (x = 0; x < ARRAY_LEN(user->prefs.order); x++) {
		format_t format = ast_codec_pref_index(&user->prefs, x);
		const char *skype_codec = NULL;
		
		if (!format) {
			break;
		}
		
		switch (format) {
		case AST_FORMAT_ULAW:
			skype_codec = "PCMU";
			break;
		case AST_FORMAT_ALAW:
			skype_codec = "PCMA";
			break;
		case AST_FORMAT_G729A:
			skype_codec = "G729";
			break;
		default:
			ast_log(LOG_WARNING, "Codec '%s' not supported by Skype for Asterisk (user '%s'). Ignoring.\n", ast_getformatname(format), user->name);
			break;
		}
		
		if (skype_codec) {
			ast_build_string(&codec_str, &codec_buf_size, "%s:", skype_codec);
		}
	}
	
	sfa_user_set_codecs(user->sfa_user, codec_buf);
	
	sfa_user_set_buddy_state_updates(user->sfa_user, user->buddy_presence ? SFA_BUDDY_STATE_UPDATE_ALL : SFA_BUDDY_STATE_UPDATE_NONE);
}

static int create_sfa_user(struct client_user_info *user) {
	if (user->sfa_user) {
		return 0;
	}

	if (!(user->sfa_user = sfa_user_new(user->name, user, USER_DEBUG(user), USER_VEDEBUG(user)))) {
		ast_log(LOG_ERROR, "Unable to connect user %s to Skype network\n", user->name);
		return -1;
	}

	if (option_verbose > 2) {
		ast_verbose(VERBOSE_PREFIX_3 "Launched Skype for username '%s'\n", user->name);
	}

	return 0;
}

/* Return a new, initialized user */
static struct client_user_info *create_user(const char *name)
{
	struct client_user_info *user;

	if (!(user = ao2_alloc(sizeof(*user), destroy_client_user))) {
		return NULL;
	}

	if (ast_string_field_init(user, 256)) {
		ao2_ref(user, -1);
		return NULL;
	}

	AST_LIST_HEAD_INIT_NOLOCK(&user->auth_policies);

	if (!(user->calls = ao2_container_alloc(1, call_hash_cb, NULL))) {
		ao2_ref(user, -1);
		return NULL;
	}

	if (!ast_strlen_zero(name)) {
		ast_string_field_set(user, name, name);
	}

	return user;
}

#define UPDATE_SFA_PROPERTY(__u,__field,__var,__val) \
	ast_string_field_set(__u, __field, __val); \
	if (__u->sfa_user) { \
		sfa_account_set_property(__u->sfa_user, sfa_str2property(__var), __val, 0); \
		__u->property_update_queued = 1; \
	}

static void handle_user_option(struct client_user_info *user, const char *var, const char *val)
{
	if (!strcasecmp(var, "debug") && user->debug != ast_true(val)) {
		user->debug = ast_true(val);
		if (user->sfa_user) {
			sfa_user_set_debug(user->sfa_user, USER_DEBUG(user));
		}
	} else if (!strcasecmp(var, "vedebug") && user->vedebug != ast_true(val)) {
		user->vedebug = ast_true(val);
		if (user->sfa_user) {
			sfa_user_set_debug(user->sfa_user, USER_VEDEBUG(user));
		}
	} else if (!strcasecmp(var, "secret") && strcmp(user->secret, val)) {
		ast_string_field_set(user, secret, val);
		if (user->sfa_user) {
			user->login_queued = 1;
		}
	} else if (!strcasecmp(var, "context")) {
		ast_string_field_set(user, context, val);
	} else if (!strcasecmp(var, "exten")) {
		ast_string_field_set(user, exten, val);
	} else if (!strcasecmp(var, "accountcode")) {
		ast_string_field_set(user, accountcode, val);
	} else if (!strcasecmp(var, "allow") && !ast_strlen_zero(val)) {
		ast_parse_allow_disallow(&user->prefs, &user->capability, val, 1);
	} else if (!strcasecmp(var, "disallow") && !ast_strlen_zero(val)) {
		ast_parse_allow_disallow(&user->prefs, &user->capability, val, 0);
	} else if (!strcasecmp(var, "buddy_autoadd")) {
		set_buddy_autoadd_flags(user, val);
	} else if (!strcasecmp(var, "buddy_presence") && user->buddy_presence != !ast_false(val)) {
		user->buddy_presence = !ast_false(val);
		if (user->sfa_user) {
		sfa_user_set_buddy_state_updates(user->sfa_user,
			user->buddy_presence ? SFA_BUDDY_STATE_UPDATE_ALL : SFA_BUDDY_STATE_UPDATE_NONE);
		}
	} else if (!strcasecmp(var, "direction")) {
		if (!strcasecmp(val, "both")) {
			user->direction = DIRECTION_BOTH;
		} else if (!strcasecmp(val, "incoming")) {
			user->direction = DIRECTION_INCOMING;
		} else if (!strcasecmp(val, "outgoing")) {
			user->direction = DIRECTION_OUTGOING;
		} else {
			ast_log(LOG_WARNING, "Invalid 'direction' value '%s' for user '%s'\n", val, user->name);
		}
	} else if (!strcasecmp(var, "amaflags")) {
		int format;

		if ((format = ast_cdr_amaflags2int(val)) < 0) {
			ast_log(LOG_WARNING, "Invalid AMA Flags: %s for user %s\n", val, user->name);
		} else {
			user->amaflags = format;
		}
	} else if (!strcasecmp(var, "mohinterpret")) {
		ast_string_field_set(user, mohinterpret, val);
	} else if (!strcasecmp(var, "mohsuggest")) {
		ast_string_field_set(user, mohsuggest, val);
	} else if (!strcasecmp(var, "auth_policy")) {
		if (add_user_auth_policy(user, val)) {
			ast_log(LOG_WARNING, "Invalid 'authpolicy' value '%s' for user '%s'\n", val, user->name);
		}
	} else if (!strcasecmp(var, "language") && strcmp(user->language, val)) {
		UPDATE_SFA_PROPERTY(user, language, var, val);
	} else if (!strcasecmp(var, "fullname") && strcmp(user->fullname, val)) {
		UPDATE_SFA_PROPERTY(user, fullname, var, val);
	} else if (!strcasecmp(var, "country") && strcmp(user->country, val)) {
		UPDATE_SFA_PROPERTY(user, country, var, val);
	} else if (!strcasecmp(var, "province") && strcmp(user->province, val)) {
		UPDATE_SFA_PROPERTY(user, province, var, val);
	} else if (!strcasecmp(var, "city") && strcmp(user->city, val)) {
		UPDATE_SFA_PROPERTY(user, city, var, val);
	} else if (!strcasecmp(var, "phone_home") && strcmp(user->phone_home, val)) {
		UPDATE_SFA_PROPERTY(user, phone_home, var, val);
	} else if (!strcasecmp(var, "phone_office") && strcmp(user->phone_office, val)) {
		UPDATE_SFA_PROPERTY(user, phone_office, var, val);
	} else if (!strcasecmp(var, "phone_mobile") && strcmp(user->phone_mobile, val)) {
		UPDATE_SFA_PROPERTY(user, phone_mobile, var, val);
	} else if (!strcasecmp(var, "email") && strcmp(user->email, val)) {
		UPDATE_SFA_PROPERTY(user, email, var, val);
	} else if (!strcasecmp(var, "homepage") && strcmp(user->homepage, val)) {
		UPDATE_SFA_PROPERTY(user, homepage, var, val);
	} else if (!strcasecmp(var, "about") && strcmp(user->about, val)) {
		UPDATE_SFA_PROPERTY(user, about, var, val);
	} else if (!strcasecmp(var, "avatar") && strcmp(user->avatar_filename, val)) {
		UPDATE_SFA_PROPERTY(user, avatar_filename, var, val);
	} else if (!strcasecmp(var, "birthday") && strcmp(user->birthday, val)) {
		UPDATE_SFA_PROPERTY(user, birthday, var, val);
	} else if (!strcasecmp(var, "gender") && strcmp(user->gender, val)) {
		UPDATE_SFA_PROPERTY(user, gender, var, val);
	} else if (!strcasecmp(var, "timezone") && strcmp(user->timezone, val)) {
		UPDATE_SFA_PROPERTY(user, timezone, var, val);
	} else if (!strcasecmp(var, "mood_text") && strcmp(user->mood_text, val)) {
		UPDATE_SFA_PROPERTY(user, mood_text, var, val);
	} else if (!strcasecmp(var, "autoreply") && strcmp(user->autoreply, val)) {
		UPDATE_SFA_PROPERTY(user, autoreply, var, val);
	} else if (!strcasecmp(var, "webpresence") && ast_true(user->webpresence) != ast_true(val)) {
		UPDATE_SFA_PROPERTY(user, webpresence, var, val);
	}
}

static int options_cleanup(void *obj, void *arg, int flags)
{
	struct config_option *opt = obj;
	struct client_user_info *user = arg;

	if (!opt->parsed) {
		handle_user_option(user, opt->name, opt->default_value);
	}

	return CMP_MATCH;
}

static struct client_user_info *build_user(const char *cat, struct ast_variable *vars, int reload)
{
	struct client_user_info *user;
	int add_user = 0;
	struct ast_variable *v;
	struct ao2_container *user_options;

	if (!(user_options = ao2_container_alloc(53, option_hash_cb, option_cmp_cb))) {
		return NULL;
	}

	if (!(user = find_user(cat))) {
		if (!(user = create_user(cat))) {
			return NULL;
		}
		ao2_link(users, user);
		if (reload) {
			add_user = 1;
		}
	}

	if (reload) {
		user->fresh = 1;
	}

	INIT_CONFIG_OPTION(user_options, "debug", "false");
	INIT_CONFIG_OPTION(user_options, "vedebug", "false");
	INIT_CONFIG_OPTION(user_options, "secret", "");
	INIT_CONFIG_OPTION(user_options, "context", "default");
	INIT_CONFIG_OPTION(user_options, "exten", user->name);
	INIT_CONFIG_OPTION(user_options, "accountcode", "");
	INIT_CONFIG_OPTION(user_options, "disallow", "");
	INIT_CONFIG_OPTION(user_options, "allow", "ulaw,g729");
	INIT_CONFIG_OPTION(user_options, "buddy_autoadd", "false");
	INIT_CONFIG_OPTION(user_options, "buddy_presence", "true");
	INIT_CONFIG_OPTION(user_options, "direction", "both");
	INIT_CONFIG_OPTION(user_options, "amaflags", "default");
	INIT_CONFIG_OPTION(user_options, "mohinterpret", "default");
	INIT_CONFIG_OPTION(user_options, "mohsuggest", "default");
	INIT_CONFIG_OPTION(user_options, "auth_policy", "accept");
	INIT_CONFIG_OPTION(user_options, "language", "");
	INIT_CONFIG_OPTION(user_options, "fullname", "");
	INIT_CONFIG_OPTION(user_options, "country", "");
	INIT_CONFIG_OPTION(user_options, "province", "");
	INIT_CONFIG_OPTION(user_options, "city", "");
	INIT_CONFIG_OPTION(user_options, "phone_home", "");
	INIT_CONFIG_OPTION(user_options, "phone_office", "");
	INIT_CONFIG_OPTION(user_options, "phone_mobile", "");
	INIT_CONFIG_OPTION(user_options, "email", "");
	INIT_CONFIG_OPTION(user_options, "homepage", "");
	INIT_CONFIG_OPTION(user_options, "about", "");
	INIT_CONFIG_OPTION(user_options, "avatar", "");
	INIT_CONFIG_OPTION(user_options, "birthday", "");
	INIT_CONFIG_OPTION(user_options, "gender", "");
	INIT_CONFIG_OPTION(user_options, "timezone", "");
	INIT_CONFIG_OPTION(user_options, "mood_text", "");
	INIT_CONFIG_OPTION(user_options, "autoreply", "");
	INIT_CONFIG_OPTION(user_options, "webpresence", "");

	ao2_lock(user);

	/* If we are an existing user, go ahead and delete existing auth_policies. It doesn't
	 * matter if they've changed or not, they'll be correct after we're done. */
	if (!add_user) {
		struct auth_policy *pol;

		/* We are found in the new config, don't delete us! */
		user->deletepending = 0;

		while ((pol = AST_LIST_REMOVE_HEAD(&user->auth_policies, next))) {
			ast_free(pol);
		}
	}

	for (v = vars; v; v = v->next) {
		struct config_option *opt;
		if (!(opt = find_option(user_options, v->name))) {
			ast_log(LOG_WARNING, "Unknown config option %s for user %s, line %u\n", v->name, user->name, v->lineno);
			continue;
		}
		handle_user_option(user, v->name, v->value);
		opt->parsed = 1;
	}

	/* Go through and unlink all options, but call handle_user_option on any that weren't
	 * parsed with the default values so that we reset options that have been commented out */
	ao2_callback(user_options, OBJ_MULTIPLE | OBJ_NODATA | OBJ_UNLINK, options_cleanup, user);
	ao2_ref(user_options, -1);
	
	if (add_user) {
		create_sfa_user(user);
	}

	if (user->property_update_queued) {
		sfa_account_commit_properties(user->sfa_user);
		user->property_update_queued = 0;
	}

	if (user->login_queued || add_user) {
		connect_user(user);
		user->login_queued = 0;
	}

	ao2_unlock(user);

	return user;
}

static void connect_users(void)
{
	struct ao2_iterator i;
	struct client_user_info *user;

	i = ao2_iterator_init(users, 0);

	while ((user = ao2_iterator_next(&i))) {
		ao2_lock(user);

		if (create_sfa_user(user)) {
			ast_log(LOG_ERROR, "Unable to connect user %s to Skype network\n", user->name);
			ao2_unlock(user);
			ao2_ref(user, -1);
			continue;
		}

		connect_user(user);

		ao2_unlock(user);
		ao2_ref(user, -1);
	}

	if (ao2_iterator_destroy != NULL) {
		ao2_iterator_destroy(&i);
	}
}

static char *show_settings_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "skype show settings";
		e->usage =
			"Usage: skype show settings\n"
			"       Show settings (from [general] section of chan_skype.conf).\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	ast_cli(a->fd, "Skype For Asterisk Settings:\n");
	ast_cli(a->fd, "\tengine_directory: %s\n", global_config.engine_directory);
	ast_cli(a->fd, "\tdata_directory: %s\n", global_config.data_directory);
	ast_cli(a->fd, "\tdefaultuser: %s\n", global_config.defaultuser);
	ast_cli(a->fd, "\tbind_address: %s\n", global_config.bind_address);
	ast_cli(a->fd, "\tbind_port: %s\n", global_config.bind_port);
	ast_cli(a->fd, "\trtp_address: %s\n", global_config.rtp_address);
	ast_cli(a->fd, "\thttps_proxy: %s\n", global_config.https_proxy);
	ast_cli(a->fd, "\thttps_proxy_user: %s\n", global_config.https_proxy_user);
	ast_cli(a->fd, "\thttps_proxy_password: %s\n", global_config.https_proxy_password);
	ast_cli(a->fd, "\tsocks5_proxy: %s\n", global_config.socks5_proxy);
	ast_cli(a->fd, "\tsocks5_proxy_user: %s\n", global_config.socks5_proxy_user);
	ast_cli(a->fd, "\tsocks5_proxy_password: %s\n", global_config.socks5_proxy_password);
	ast_cli(a->fd, "\tdisable_tcpauto: %s\n", global_config.disable_tcpauto ? "yes" : "no");
	ast_cli(a->fd, "\tdisable_udp: %s\n", global_config.disable_udp ? "yes" : "no");
	ast_cli(a->fd, "\tdebug: %s\n", global_config.debug ? "yes" : "no");

	return CLI_SUCCESS;
}

static char *show_version_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch(cmd) {
	case CLI_INIT:
		e->command = "skype show version";
		e->usage =
			"Usage: skype show version\n"
			"       Show versions of Skype For Asterisk components.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	ast_cli(a->fd, "Skype For Asterisk Components:\n");
	ast_cli(a->fd, "\tChannel Driver: " SFA_VERSION "\n");
	ast_cli(a->fd, "\tLibrary: %s\n", sfa_get_version());

	return CLI_SUCCESS;
}

static char *show_users_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct client_user_info *user;

	switch(cmd) {
	case CLI_INIT:
		e->command = "skype show users";
		e->usage = 
			"Usage: skype show users\n"
			"       List all configured Skype users.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	if (!users) {
		ast_log(LOG_ERROR, "Skype for Asterisk could not be loaded\n");
		ast_log(LOG_ERROR, "Exit Asterisk and check for hung 'skypeforasterisk' processes\n");
		return CLI_SUCCESS;
	}

	ast_cli(a->fd, "Skype Users\n");

	for (i = ao2_iterator_init(users, 0), user = ao2_iterator_next(&i);
	     user;
	     ao2_ref(user, -1), user = ao2_iterator_next(&i)) {
		/* no need to lock the user here, only non-changing data is being accessed */
		ast_cli(a->fd, "%s: %s\n", user->name, user->sfa_user ? sfa_user_account_status(user->sfa_user) : "Connection Error");
	}

	if (ao2_iterator_destroy != NULL) {
		ao2_iterator_destroy(&i);
	}

	return CLI_SUCCESS;
}

static char *complete_skype_user(const char *word, int state)
{
	int which = 0;
	struct client_user_info *user;
	char *res = NULL;
	int wordlen = strlen(word);
	struct ao2_iterator i;

	if (!users) {
		return NULL;
	}

	for (i = ao2_iterator_init(users, 0), user = ao2_iterator_next(&i);
	     user;
	     ao2_ref(user, -1), user = ao2_iterator_next(&i)) {
		/* no need to lock the user here, only non-changing data is being accessed */
		if (!strncasecmp(user->name, word, wordlen) && (++which > state)) {
			res = ast_strdup(user->name);
			ao2_ref(user, -1);
			break;
		}
	}

	if (ao2_iterator_destroy != NULL) {
		ao2_iterator_destroy(&i);
	}

	return res;
}

static char *set_debug_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int debug, vedebug;

	switch (cmd) {
	case CLI_INIT:
		e->command = "skype set {debug|vedebug} {on|off}";
		e->usage =
			"Usage: skype set {debug|vedebug} {on|off} (user, username)\n"
			"       Enables Skype or voicengine debugging\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 4) {
			return a->n == 0 ? ast_strdup("user") : NULL;
		} else if (a->pos > 4 && strcasestr(a->line, " user")) {
			return complete_skype_user(a->word, a->n);
		}
		return NULL;
	}
	if (a->argc < 4) {
		return CLI_SHOWUSAGE;
	}

	debug = ast_true(a->argv[3]);

	vedebug = !strcasecmp(a->argv[2], "vedebug");

	if (a->argc == 4) {
		debug = ast_true(a->argv[3]);
		if (!vedebug) {
			global_config.debug = debug;
			sfa_set_global_debug(GLOBAL_DEBUG);
		} else {
			ast_cli(a->fd, "Skype global Voice Engine debugging not supported.");
			return CLI_FAILURE;
		}
		ast_cli(a->fd, "Skype global debugging %s\n", debug ? "enabled" : "disabled");
	} else {
		unsigned int x;
		struct client_user_info *user;

		for (x = 5; x < a->argc; x++) {
			if ((user = find_active_user(a->argv[x]))) {
				/* no need to lock the user object here; the fields we access either
				   cannot change (name) or we are writing to them so we don't care
				*/
				if (!vedebug) {
					user->debug = debug;
					sfa_user_set_debug(user->sfa_user, USER_DEBUG(user));
				} else {
					user->vedebug = debug;
					sfa_user_set_vedebug(user->sfa_user, USER_VEDEBUG(user));
				}
				ast_cli(a->fd, "%sDebugging for Skype user '%s' %s\n", vedebug ? "Voice Engine " : "", user->name,
					debug ? "enabled" : "disabled");
				ao2_ref(user, -1);
			} else {
				ast_cli(a->fd, "Active Skype user '%s' not found.\n", a->argv[x]);
			}
		}
	}

	return CLI_SUCCESS;
}

static const char *buddy_autoadd2str(struct client_user_info *user)
{
	if (!user->autoadd_hint_contacts) {
		return user->autoadd_buddy_contacts ? "buddies" : "no";
	}

	return user->autoadd_buddy_contacts ? "buddies,hints" : "hints";
}

static char *show_user_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	unsigned int x;
	struct client_user_info *user;
	char outbuf[1024];

	switch(cmd) {
	case CLI_INIT:
		e->command = "skype show user";
		e->usage =
			"Usage: skype show user\n"
			"Display details about one or more configured Skype users\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3 && strcasestr(a->line, " user")) {
			return complete_skype_user(a->word, a->n);
		}
		return NULL;
	}
	if (a->argc < 4) {
		return CLI_SHOWUSAGE;
	}
	for (x = 3; x < a->argc; x++) {
		if ((user = find_user(a->argv[x]))) {
			char *build = outbuf;
			size_t left = sizeof(outbuf);
			struct ast_codec_pref pref;
			unsigned int x;
			struct auth_policy *pol;

			ast_build_string(&build, &left, "Skype User %s:\n", user->name);

			ao2_lock(user);
			ast_build_string(&build, &left, "\tsecret: %s\n", user->secret);
			ast_build_string(&build, &left, "\tcontext: %s\n", user->context);
			ast_build_string(&build, &left, "\texten: %s\n", user->exten);
			ast_build_string(&build, &left, "\taccountcode: %s\n", user->accountcode);
			ast_build_string(&build, &left, "\tlanguage: %s\n", user->language);
			ast_build_string(&build, &left, "\tmohinterpret: %s\n", user->mohinterpret);
			ast_build_string(&build, &left, "\tmohsuggest: %s\n", user->mohsuggest);
			ast_build_string(&build, &left, "\tamaflags: %s\n", ast_cdr_flags2str(user->amaflags));
			ast_build_string(&build, &left, "\tdirection: %s\n", user->direction == DIRECTION_BOTH ? "both" : (user->direction == DIRECTION_INCOMING ? "incoming" : "outgoing"));
			ast_build_string(&build, &left, "\tbuddy_autoadd: %s\n", buddy_autoadd2str(user));
			ast_build_string(&build, &left, "\tbuddy_presence: %s\n", user->buddy_presence ? "yes" : "no");
			ast_build_string(&build, &left, "\tdebug: %s\n", user->debug ? "yes" : "no");
			ast_build_string(&build, &left, "\tvdebug: %s\n", user->vedebug ? "yes" : "no");
			pref = user->prefs;
			if (AST_LIST_EMPTY(&user->auth_policies)) {
				ast_build_string(&build, &left, "\tauth_policy: accept\n");
			} else {
				ast_build_string(&build, &left, "\tauth_policy:\n");
				AST_LIST_TRAVERSE(&user->auth_policies, pol, next) {
					switch (pol->type) {
					case AUTH_ACCEPT:
						ast_build_string(&build, &left, "\t\taccept%s%s\n", ast_strlen_zero(pol->secret) ? "" : ":", pol->secret);
						break;
					case AUTH_DENY:
						ast_build_string(&build, &left, "\t\tdeny\n");
						break;
					case AUTH_BLOCK:
						ast_build_string(&build, &left, "\t\tblock\n");
						break;
					case AUTH_IGNORE:
						ast_build_string(&build, &left, "\t\tignore\n");
						break;
					default:
						break;
					}
				}
			}
			ao2_unlock(user);
			ast_build_string(&build, &left, "\tcodecs: ");
			for (x = 0; x < ARRAY_LEN(pref.order); x++) {
				int format = ast_codec_pref_index(&pref, x);

				if (!format) {
					break;
				}

				ast_build_string(&build, &left, "%s,", ast_getformatname(format));
			}
			build--;
			left++;
			ast_build_string(&build, &left, "\n");
			ao2_ref(user, -1);
			ast_cli(a->fd, "%s", outbuf);
		} else {
			ast_cli(a->fd, "Skype user '%s' not found.\n", a->argv[x]);
		}
	}

	return CLI_SUCCESS;
}

static void buddy_cli_write(struct client_user_info *user, const char *name, const char *fullname, void *data)
{
	struct ast_cli_args *a = data;
	enum sfa_buddy_state state;
   
	if (!user->sfa_user) {
		ast_log(LOG_WARNING, "%s was unable to connect to the Skype network\n", user->name);
		return;
	}

	state = sfa_user_get_buddy_state(user->sfa_user, name, 0);
	ast_cli(a->fd, "%-20.20s: %s\n", name, sfa_buddy_state2str(state));
}

static char *show_buddies_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct client_user_info *user;

	switch(cmd) {
	case CLI_INIT:
		e->command = "skype show buddies";
		e->usage =
			"Usage: skype show buddies <user>\n"
			"Display a Skype user's buddy list\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3 && strcasestr(a->line, " buddies")) {
			return complete_skype_user(a->word, a->n);
		}
		return NULL;
	}
	if (a->argc < 4) {
		return CLI_SHOWUSAGE;
	}

	if (!(user = find_active_user(a->argv[3]))) {
		ast_cli(a->fd, "Active Skype user '%s' not found.\n", a->argv[3]);
		return CLI_SUCCESS;
	}

	ast_cli(a->fd, "Contact list for %s\n", user->name);
	sfa_buddylist_iterate(user->sfa_user, buddy_cli_write, a);

	ao2_ref(user, -1);

	return CLI_SUCCESS;
}

static char *show_hostid_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "skype show hostid";
		e->usage =
			"Usage: skype show hostid\n"
			"Display Skype for Asterisk Host-ID\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	ast_cli(a->fd, "Host-ID: %s\n", sfa_get_hostid());

	return CLI_SUCCESS;
}

static char *show_licenses_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	const struct sfa_license *license;
	struct sfa_license_iterator *iter;

	switch (cmd) {
	case CLI_INIT:
		e->command = "skype show licenses";
		e->usage =
			"Usage: skype show licenses\n"
			"Display Skype for Asterisk licenses\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	ast_cli(a->fd, "%s", "Skype For Asterisk Licensing Information\n");
	ast_cli(a->fd, "%s", "========================================\n");
	ast_cli(a->fd, "Total licensed channels: %d\n", sfa_get_maxcalls());

	if (!(iter = sfa_license_iterator_new())) {
		ast_cli(a->fd, "\nNo Licenses Found\n");
		return CLI_SUCCESS;
	}

	ast_cli(a->fd, "\nLicenses Found:\n");
	while ((license = sfa_license_iterator_next(iter))) {
		ast_cli(a->fd, "File: %s -- Key: %s -- Expires: %s -- Host-ID: %s -- Channels: %d (%s)\n",
			sfa_license_filename(license), sfa_license_key(license),
			sfa_license_expires(license), sfa_license_hostid(license),
			sfa_license_channels(license), sfa_license_status(license));
	}

	sfa_license_iterator_destroy(iter);

	return CLI_SUCCESS;
}

static char *logout_user_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct client_user_info *user;

	switch(cmd) {
	case CLI_INIT:
		e->command = "skype logout user";
		e->usage =
			"Usage: skype logout user <user\n"
			"Log a Skype user out\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3 && strcasestr(a->line, " user")) {
			return complete_skype_user(a->word, a->n);
		}
		return NULL;
	}

	if (a->argc < 4) {
		return CLI_SHOWUSAGE;
	}

	if (!(user = find_active_user(a->argv[3]))) {
		ast_cli(a->fd, "Active Skype user '%s' not found.\n", a->argv[3]);
		return CLI_SUCCESS;
	}

	if (!sfa_user_loggedin(user->sfa_user)) {
		ast_cli(a->fd, "Skype user '%s' not logged in.\n", a->argv[3]);
		ao2_ref(user, -1);
		return CLI_SUCCESS;
	}

	sfa_user_logout(user->sfa_user);
	ao2_ref(user, -1);

	return CLI_SUCCESS;
}

static char *login_user_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct client_user_info *user;

	switch(cmd) {
	case CLI_INIT:
		e->command = "skype login user";
		e->usage =
			"Usage: skype login user <user\n"
			"Log a Skype user in\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3 && strcasestr(a->line, " user")) {
			return complete_skype_user(a->word, a->n);
		}
		return NULL;
	}

	if (a->argc < 4) {
		return CLI_SHOWUSAGE;
	}

	if (!(user = find_active_user(a->argv[3]))) {
		ast_cli(a->fd, "Active Skype user '%s' not found.\n", a->argv[3]);
		return CLI_SUCCESS;
	}

	if (sfa_user_loggedin(user->sfa_user)) {
		ast_cli(a->fd, "Skype user '%s' already logged in.\n", a->argv[3]);
		ao2_ref(user, -1);
		return CLI_SUCCESS;
	}

	sfa_user_login(user->sfa_user, user->secret);
	ao2_ref(user, -1);

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli[] = {
	AST_CLI_DEFINE(show_settings_cli, "Show general settings for Skype for Asterisk"),
	AST_CLI_DEFINE(show_version_cli, "Show versions of Skype For Asterisk components"),
	AST_CLI_DEFINE(show_users_cli, "List configured Skype users"),  
	AST_CLI_DEFINE(show_user_cli, "Display details about one or more configured Skype users"),
	AST_CLI_DEFINE(show_hostid_cli, "Display the Skype for Asterisk Host-ID"),
	AST_CLI_DEFINE(show_licenses_cli, "Display Skype for Asterisk licenses"),
	AST_CLI_DEFINE(set_debug_cli, "Enable/disable Skype/Voice Engine debugging for Skype channels"), 
	AST_CLI_DEFINE(show_buddies_cli, "Display a Skype user's buddies"),
	AST_CLI_DEFINE(logout_user_cli, "Log out a Skype user"),
	AST_CLI_DEFINE(login_user_cli, "Log in a Skype user"),
};

static struct client_user_info *get_manager_user(const struct message *m, const char **errmsg)
{
	const char *user_header;
	struct client_user_info *user;

	user_header = astman_get_header(m, "User");
	if (ast_strlen_zero(user_header)) {
		*errmsg = "User: <name> missing.";
		return NULL;
	}

	if (!(user = find_active_user(user_header))) {
		*errmsg = "User not found or not active.";
		return NULL;
	}

	return user;
}

struct sb_info {
	struct mansession *s;
	char idtext[256];
	size_t total;
};

static void buddy_man_write(struct client_user_info *user, const char *skypename, const char *fullname, void *data)
{
	struct sb_info *sb = data;
	enum sfa_buddy_state state;
	
	if (!user->sfa_user) {
		ast_log(LOG_WARNING, "%s was unable to connect to the Skype network\n", user->name);
		return;
	}

	state = sfa_user_get_buddy_state(user->sfa_user, skypename, 0);

	sb->total++;
	astman_append(sb->s,
		"Event: SkypeBuddyEntry\r\n"
		"%s"
		"Buddy: %s\r\n"
		"Status: %s\r\n"
		"Fullname: %s\r\n"
		"\r\n",
		sb->idtext, skypename, sfa_buddy_state2str(state), fullname
	);
}

static int show_buddies_manager(struct mansession *s, const struct message *m)
{
	struct client_user_info *user;
	const char *errmsg = NULL;
	const char *id = astman_get_header(m, "ActionID");
	struct sb_info sb = {
		.s = s,
		.total = 0,
	};

	if (!(user = get_manager_user(m, &errmsg))) {
		astman_send_error(s, m, (char *) errmsg);
		return 0;
	}

	astman_send_ack(s, m, "Skype buddy status list will follow");

	if (!ast_strlen_zero(id)) {
		snprintf(sb.idtext, sizeof(sb.idtext), "ActionID: %s\r\n", id);
	}
	
	sfa_buddylist_iterate(user->sfa_user, buddy_man_write, &sb);

	astman_append(s,
		      "Event: SkypeBuddylistComplete\r\n"
		      "ListItems: %zu\r\n"
		      "%s"
		      "\r\n",
		      sb.total, sb.idtext
	);

	ao2_ref(user, -1);

	return 0;
}

static int add_buddy_manager(struct mansession *s, const struct message *m)
{
	struct client_user_info *user;
	const char *errmsg, *buddyname, *authmsg;

	if (!(user = get_manager_user(m, &errmsg))) {
		astman_send_error(s, m, (char *) errmsg);
		return 0;
	}

	buddyname = astman_get_header(m, "Buddy");
	if (ast_strlen_zero(buddyname)) {
		astman_send_error(s, m, "Buddy: <name> missing.");
		ao2_ref(user, -1);
		return 0;
	}

	if (sfa_add_contact(user->sfa_user, buddyname)) {
		astman_send_error(s, m, "Could not add buddy");
		ao2_ref(user, -1);
		return 0;
	}

	authmsg = astman_get_header(m, "AuthMsg");
	if (sfa_request_contact_auth(user->sfa_user, buddyname, authmsg)) {
		astman_send_error(s, m, "Could not send buddy auth request");
		ao2_ref(user, -1);
		return 0;
	}

	astman_send_ack(s, m, "Contact added");

	ao2_ref(user, -1);
	return 0;
}

static int remove_buddy_manager(struct mansession *s, const struct message *m)
{
	struct client_user_info *user;
	const char *errmsg, *buddyname;

	if (!(user = get_manager_user(m, &errmsg))) {
		astman_send_error(s, m, (char *) errmsg);
		return 0;
	}

	buddyname = astman_get_header(m, "Buddy");
	if (ast_strlen_zero(buddyname)) {
		astman_send_error(s, m, "Buddy: <name> missing.");
		ao2_ref(user, -1);
		return 0;
	}

	if (sfa_remove_contact(user->sfa_user, buddyname)) {
		astman_send_error(s, m, "Could not remove buddy");
		ao2_ref(user, -1);
		return 0;
	}

	astman_send_ack(s, m, "Contact removed");
	ao2_ref(user, -1);
	return 0;
}

static void manager_show_buddy_write(struct client_user_info *user, const char *property, const char *value, void *data)
{
	struct mansession *s = data;

	if (ast_strlen_zero(property) || ast_strlen_zero(value) || !strcasecmp(property, "avatar")) {
		return;
	}

	if (!strcasecmp(property, "availability")) {
		astman_append(s, "%s: %s\r\n", property, sfa_buddy_state2str(atoi(value)));
	} else {
		astman_append(s, "%s: %s\r\n", property, value);
	}
}

static int show_buddy_manager(struct mansession *s, const struct message *m)
{
	struct client_user_info *user;
	const char *id, *buddy, *errmsg;

	if (!(user = get_manager_user(m, &errmsg))) {
		astman_send_error(s, m, (char *) errmsg);
		return 0;
	}

	buddy = astman_get_header(m, "Buddy");
	if (ast_strlen_zero(buddy)) {
		astman_send_error(s, m, "Buddy: <name> missing.");
		ao2_ref(user, -1);
		return 0;
	}

	astman_append(s, "Response: Success\r\n");
	id = astman_get_header(m, "ActionID");
	if (!ast_strlen_zero(id)) {
		astman_append(s, "ActionID: %s\r\n", id);
	}

	sfa_buddy_property_iterate(user->sfa_user, buddy, manager_show_buddy_write, s);

	astman_append(s, "\r\n");
	ao2_ref(user, -1);
	return 0;
}

static int account_property_write_manager(struct mansession *s, const struct message *m)
{
	struct client_user_info *user;
	const char *errmsg;
	struct ast_variable *properties, *var;

	if (!(user = get_manager_user(m, &errmsg))) {
		astman_send_error(s, m, (char *) errmsg);
		return 0;
	}

	if (!(properties = astman_get_variables(m))) {
		astman_send_error(s, m, "No variable properties found.\n");
		ao2_ref(user, -1);
		return 0;
	}

	for (var = properties; var; var = var->next) {
		enum sfa_property prop;
		if ((prop = sfa_str2property(var->name))) {
			sfa_account_set_property(user->sfa_user, prop, var->value, 0);
		}
	}

	sfa_account_commit_properties(user->sfa_user);
	astman_send_ack(s, m, "Properties set");

	ast_variables_destroy(properties);
	ao2_ref(user, -1);

	return 0;
}

static int license_status_manager(struct mansession *s, const struct message *m)
{
	const char *id = astman_get_header(m, "ActionID");
	char idtext[256] = "";

	if (!(ast_strlen_zero(id))) {
		snprintf(idtext, sizeof(idtext), "ActionID: %s\r\n", id);
	}

	astman_append(s,
		"Response: Success\r\n"
		"%s"
		"CallsLicensed: %d\r\n"
		"\r\n",
		idtext, sfa_get_maxcalls());

	return 0;
}

static int license_list_manager(struct mansession *s, const struct message *m)
{
	const char *id = astman_get_header(m, "ActionID");
	char idtext[256] = "";
	const struct sfa_license *license;
	struct sfa_license_iterator *iter;

	if (!ast_strlen_zero(id)) {
		snprintf(idtext, sizeof(idtext), "ActionID: %s\r\n", id);
	}

	astman_send_ack(s, m, "License list will follow");

	if (!(iter = sfa_license_iterator_new())) {
		goto licenselist_end;
	}

	while ((license = sfa_license_iterator_next(iter))) {
		astman_append(s,
			"Event: SkypeLicense\r\n"
			"%s"
			"File: %s\r\n"
			"Key: %s\r\n"
			"Expires: %s\r\n"
			"Host-ID: %s\r\n"
			"Channels: %d\r\n"
			"Status: %s\r\n"
			"\r\n",
			idtext,
			sfa_license_filename(license), sfa_license_key(license),
			sfa_license_expires(license), sfa_license_hostid(license),
			sfa_license_channels(license), sfa_license_status(license));
	}

	sfa_license_iterator_destroy(iter);

licenselist_end:
	astman_append(s,
		"Event: SkypeLicenseList complete\r\n"
		"%s"
		"\r\n", idtext);
	return 0;
}

static int send_chat_manager(struct mansession *s, const struct message *m)
{
	const char *errmsg, *toname, *message;
	struct client_user_info *user;
	const char *encoding;

	if (!(user = get_manager_user(m, &errmsg))) {
		astman_send_error(s, m, (char *) errmsg);
		return 0;
	}

	if (!(toname = astman_get_header(m, "Skypename"))) {
		astman_send_error(s, m, "Must provide a skypename to send the message to\n");
		return 0;
	}

	if (!(message = astman_get_header(m, "Message"))) {
		astman_send_error(s, m, "Must provide a message to send!\n");
		return 0;
	}

	if ((encoding = astman_get_header(m, "Encoding")) && !strcasecmp(encoding, "base64")) {
		size_t maxlen = strlen(message);
		unsigned char *base64 = alloca(maxlen);

		maxlen = ast_base64decode(base64, message, maxlen);
		base64[maxlen] = '\0';
		sfa_send_chat_message(user->sfa_user, toname, (const char *) base64);
	} else {
		sfa_send_chat_message(user->sfa_user, toname, message);
	}

	astman_send_ack(s, m, "Message sent");

	return 0;
}

static int skype_datastore_write(struct ast_channel *chan, const char *cmd, const char *var, const char *val, const char *dsname)
{
	struct ast_datastore *datastore;
	struct varshead *varshead;
	struct ast_var_t *v;

	if (ast_strlen_zero(var)) {
		ast_log(LOG_WARNING, "SKYPE_CALL_PROPERTY() requires a property name\n");
		return -1;
	}

	ast_channel_lock(chan);
	if (!(datastore = ast_channel_datastore_find(chan, &skype_call_datastore, dsname))) {
		if (!(datastore = ast_datastore_alloc(&skype_call_datastore, dsname))) {
			ast_log(LOG_ERROR, "Unable to allocate skype call datastore\n");
			ast_channel_unlock(chan);
			return -1;
		}

		if (!(varshead = ast_calloc(1, sizeof(*varshead)))) {
			ast_datastore_free(datastore);
			ast_channel_unlock(chan);
			return -1;
		}

		datastore->data = varshead;
		datastore->inheritance = DATASTORE_INHERIT_FOREVER;
		ast_channel_datastore_add(chan, datastore);
	}

	varshead = datastore->data;

	AST_LIST_TRAVERSE(varshead, v, entries) {
		/* Remove existing values */
		if (!strcasecmp(var, ast_var_name(v))) {
			AST_LIST_REMOVE(varshead, v, entries);
			ast_var_delete(v);
			break;
		}
	}

	v = ast_var_assign(var, S_OR(val, ""));
	AST_LIST_INSERT_HEAD(varshead, v, entries);
	ast_channel_unlock(chan);

	return 0;
}

/* Currently, only the call properties (topic and token) are writable via diaplan function
 * as setting account or contact information during a call doesn't seem useful
 */
static int acf_skype_callprop_write(struct ast_channel *chan, const char *cmd, char *var, const char *val)
{
	return skype_datastore_write(chan, cmd, var, val, "skype_call_property");
}

static struct ast_custom_function skype_callprop_function = {
	.name = "SKYPE_CALL_PROPERTY",
	.synopsis = "Set a Skype Call property",
	.syntax = "SKYPE_CALL_PROPERTY(<property>)",
	.desc =
		"Get or set a property in a skype call/caller.\n"
		"Available call properties (read/write):\n"
		"  topic\n"
		"  token\n"
		"Available caller properties (read only):\n"
		"  fullname\n"
		"  country\n"
		"  province\n"
		"  city\n"
		"  phone_home\n"
		"  phone_office\n"
		"  phone_mobile\n"
		"  email\n"
		"  homepage\n"
		"  about\n"
		"  birthday (YYYYMMDD)\n"
		"  gender (male=1, female=2)\n"
		"  target_identity\n"
		"  forwarded_by\n"
		"\n"
		"Example: Set(SKYPE_CALL_PROPERTY(topic)=Weekly Conference Call)\n"
		"         NoOp(Caller's language is ${SKYPE_CALL_PROPERTY(language)}\n",
	.read = channel_read,
	.write = acf_skype_callprop_write,
};

static int acf_skype_accountprop_read(struct ast_channel *chan, const char *cmd, char *var, char *buf, size_t len)
{
	struct client_user_info *user;
	enum sfa_property prop;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(user);
		AST_APP_ARG(property);
	);

	AST_STANDARD_APP_ARGS(args, var);

	if (ast_strlen_zero(args.user) || ast_strlen_zero(args.property)) {
		ast_log(LOG_WARNING, "%s takes both a skype user and a property\n", cmd);
		return 0;
	}

	if (!(user = find_active_user(args.user))) {
		ast_log(LOG_WARNING, "%s doesn't exist or was unable to connect to the skype network\n", args.user);
		return 0;
	}

	memset(buf, 0, len);

	if (!(prop = sfa_str2property(args.property))) {
		ast_log(LOG_WARNING, "Unrecognized property '%s'\n", args.property);
		return 0;
	}

	sfa_account_get_property(user->sfa_user, prop, buf, len);

	return 0;
}

static int acf_skype_accountprop_write(struct ast_channel *chan, const char *cmd, char *var, const char *val)
{
	struct client_user_info *user;
	enum sfa_property prop;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(user);
		AST_APP_ARG(property);
	);

	AST_STANDARD_APP_ARGS(args, var);

	if (ast_strlen_zero(args.user) || ast_strlen_zero(args.property)) {
		ast_log(LOG_WARNING, "%s takes both a skype user and a property\n", cmd);
		return 0;
	}

	if (!(user = find_active_user(args.user))) {
		ast_log(LOG_WARNING, "%s doesn't exist or was unable to connect to the skype network\n", args.user);
		return 0;
	}

	if (!(prop = sfa_str2property(args.property))) {
		ast_log(LOG_WARNING, "Unrecognized property '%s'\n", args.property);
		return 0;
	}

	sfa_account_set_property(user->sfa_user, prop, val, 0);
	sfa_account_commit_properties(user->sfa_user);

	return 0;
}

static struct ast_custom_function skype_accountprop_function = {
	.name = "SKYPE_ACCOUNT_PROPERTY",
	.synopsis = "Get a property from the called SFA account",
	.syntax = "SKYPE_ACCOUNT_PROPERTY(<account>,<property>)",
	.desc =
		"Get a property from the called SFA account.\n"
		"Available properties:\n"
		"  fullname\n"
		"  country\n"
		"  province\n"
		"  city\n"
		"  phone_home\n"
		"  phone_office\n"
		"  phone_mobile\n"
		"  email\n"
		"  homepage\n"
		"  about\n"
		"  birthday (YYYYMMDD)\n"
		"  gender (male=1, female=2)\n"
		"  status\n"
		"  availability\n"
		"\nRead only properties:\n"
		"  skypeout_currency\n"
		"  skypeout_balance\n",
		"  mood_text\n",
	.read = acf_skype_accountprop_read,
	.write = acf_skype_accountprop_write,
};

static void buddy_list_append(struct client_user_info *user, const char *name, const char *fullname, void *data)
{
	struct ast_channel *chan = data;
	enum sfa_buddy_state state;

	state = sfa_user_get_buddy_state(user->sfa_user, name, 0);
	skype_datastore_write(chan, "skype_buddies", name, sfa_buddy_state2str(state), user->name);
}

static int acf_skype_buddies(struct ast_channel *chan, const char *cmd, char *var, char *buf, size_t len)
{
	struct client_user_info *user;

	if (ast_strlen_zero(var)) {
		ast_log(LOG_WARNING, "%s requires a Skype account name\n", cmd);
		return 0;
	}

	if (!(user = find_active_user(var))) {
		ast_log(LOG_WARNING, "Active Skype user '%s' not found.\n", var);
		return 0;
	}

	sfa_buddylist_iterate(user->sfa_user, buddy_list_append, chan);

	ast_copy_string(buf, var, len);

	return 0;
}

static struct ast_custom_function skype_buddies_function = {
	.name = "SKYPE_BUDDIES",
	.synopsis = "Retrieve the list of buddies for a SFA account",
	.syntax = "SKYPE_BUDDIES(<account>)",
	.desc =
		"Returns an id to pass to SKYPE_BUDDY_FETCH(<id>) to enumerate buddies\n",
	.read = acf_skype_buddies,
};

static int acf_skype_fetch_buddy(struct ast_channel *chan, const char *cmd, char *var, char *buf, size_t len)
{
	struct varshead *varshead;
	struct ast_var_t *v;
	struct ast_datastore *datastore;

	ast_channel_lock(chan);

	if (!(datastore = ast_channel_datastore_find(chan, &skype_call_datastore, var))) {
		ast_log(LOG_WARNING, "No buddy data for '%s' stored on %s, use SKYPE_BUDDIES(%s)\n", var, chan->name, var);
		ast_channel_unlock(chan);
		return 0;
	}

	if (!(varshead = datastore->data)) {
		ast_channel_unlock(chan);
		return -1;
	}

	if ((v = AST_LIST_REMOVE_HEAD(varshead, entries))) {
		snprintf(buf, len, "%s,%s", ast_var_name(v), ast_var_value(v));
	}

	ast_channel_unlock(chan);

	return 0;
}

static struct ast_custom_function skype_buddy_fetch_function = {
	.name = "SKYPE_BUDDY_FETCH",
	.synopsis = "Retrieve the next buddy (with status) from the list stored by SKYPE_BUDDIES",
	.syntax = "SKYPE_BUDDY_FETCH(<id>)",
	.desc =
		"After calling SKYPE_BUDDIES(<account>) and obtaining a result id,\n"
		"SKYPE_BUDDY_FETCH(<id>) will return a string suitable for the ARRAY() dialplan\n"
		"funtion, containing the <buddy name>,<buddy status>\n"
		"Example:\n"
		"  Set(id=${SKYPE_BUDDIES(<account>)})\n"
		"  Set(ARRAY(buddy,status)=${SKYPE_BUDDY_FETCH(${id})})\n",
	.read = acf_skype_fetch_buddy,
};

static int acf_skype_chat_receive(struct ast_channel *chan, const char *cmd, char *var, char *buf, size_t len)
{
	char *parse = NULL;
	int timeout;
	struct timeval start;
	struct chat_subscriber *subscriber;
	struct timespec ts = {0,};
	struct timeval wait;
	int res;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(to);
		AST_APP_ARG(from);
		AST_APP_ARG(timeout);
	);

	parse = ast_strdupa(var);
	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.to) || ast_strlen_zero(args.from) || ast_strlen_zero(args.timeout)) {
		ast_log(LOG_WARNING, "%s requires arguments (to,from,timeout)\n", cmd);
		return -1;
	}

	if (sscanf(args.timeout, "%d", &timeout) != 1 || timeout < 1) {
		ast_log(LOG_WARNING, "Invalid timeout '%s'\n", args.timeout);
		return -1;
	}

	if (!(subscriber = ao2_alloc(sizeof(*subscriber), chat_subscriber_destructor))) {
		return -1;
	}

	subscriber->to = args.to;
	subscriber->from = args.from;
	ast_mutex_init(&subscriber->lock);
	ast_cond_init(&subscriber->cv, NULL);

	ao2_link(chat_subscribers, subscriber);
	start = ast_tvnow();

	if (ast_autoservice_start(chan) < 0) {
		ast_log(LOG_WARNING, "Cannot start autoservice for channel %s\n", chan->name);
		return -1;
	}

	wait = ast_tvadd(start, ast_tv(timeout, 0));
	ts.tv_sec = wait.tv_sec;
	ts.tv_nsec = wait.tv_usec * 1000;

	ast_mutex_lock(&subscriber->lock);
	res = ast_cond_timedwait(&subscriber->cv, &subscriber->lock, &ts);
	ast_mutex_unlock(&subscriber->lock);

	ao2_unlink(chat_subscribers, subscriber);

	if (res == ETIMEDOUT) {
		ast_log(LOG_DEBUG, "No message received from %s to %s in %d seconds\n", args.from, args.to, timeout);
		ao2_ref(subscriber, -1);
		ast_autoservice_stop(chan);
		return 0;
	}

	if (!ast_strlen_zero(subscriber->message)) {
		ast_copy_string(buf, subscriber->message, len);
	}

	ao2_ref(subscriber, -1);
	ast_autoservice_stop(chan);

	return 0;
}

static struct ast_custom_function skype_chat_receive_function = {
	.name = "SKYPE_CHAT_RECEIVE",
	.synopsis = "Wait for a Skype chat message from a user",
	.syntax = "SKYPE_CHAT_RECEIVE(<to>,<from>,<timeout>)",
	.desc =
		"Receives a Skype chat message sent from \"from\" and to \"to\". If no matching\n"
		"message arrives before \"timeout\", the return value will be empty.\n",
	.read = acf_skype_chat_receive,
};

static int set_sfa_globals(void)
{
	return sfa_set_globals(global_config.disable_tcpauto, global_config.disable_udp,
		&global_config.bind_sockaddr_in, &global_config.rtp_sockaddr_in, global_config.https_proxy,
		global_config.https_proxy_user, global_config.https_proxy_password, global_config.socks5_proxy,
		global_config.socks5_proxy_user, global_config.socks5_proxy_password);
}

static int cleanup_deletepending(void *obj, void *arg, int flags)
{
	struct client_user_info *user = obj;
	struct client_call_info *call;
	struct ao2_iterator i;
	
	if (!user->deletepending) {
		return 0;
	}
	
	if (!user->sfa_user) {
		return 0;
	}

	ast_verbose(VERBOSE_PREFIX_2 "Logging user %s out\n", user->name);
	/* first, hangup and cleanup any calls for this user */
	for (i = ao2_iterator_init(user->calls, 0), call = ao2_iterator_next(&i);
	     call;
	     ao2_ref(call, -1), call = ao2_iterator_next(&i)) {
		if (call->owner) {
			ast_queue_hangup(call->owner);
		}
		while (sfa_call_active(call->sfa_call)) {
			sfa_user_run(user->sfa_user, SFA_USER_RUN_BLOCK);
		}
	}
	
	if (ao2_iterator_destroy != NULL) {
		ao2_iterator_destroy(&i);
	}

	sfa_user_logout(user->sfa_user);
	while (sfa_user_loggedin(user->sfa_user)) {
		sfa_user_run(user->sfa_user, SFA_USER_RUN_BLOCK);
	}

	/* This is a blocking call */
	sfa_user_shutdown(user->sfa_user);

	return CMP_MATCH;
}

static int reload_cleanup(void *obj, void *arg, int flags)
{
	struct client_user_info *user = obj;
	
	if (!user->fresh) {
		user->deletepending = 1;
		cleanup_deletepending(obj, arg, flags);
		return CMP_MATCH;
	}

	user->fresh = 0;

	return 0;
}

static enum ast_module_load_result load_config(int reload)
{
	struct ast_config *cfg;
	struct ast_variable *v;
	const char *cat;
	struct client_user_info *user;
	struct global_config_options new_global_config = global_defaults;
	struct stat statbuf;
	struct ast_hostent ahp;
	struct hostent *hp;
	unsigned int bind_port;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };

	if (!(cfg = ast_config_load("chan_skype.conf", config_flags))) {
		/* We *must* have a config file otherwise stop immediately */
		ast_log(LOG_ERROR, "Unable to load chan_skype.conf\n");
		return AST_MODULE_LOAD_DECLINE;	
	}

	if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		ast_debug(3, "Reload, but file unchanged\n");
		return 1;
	}

	if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of chan_skype.conf are invalid and cannot be parsed\n");
		return AST_MODULE_LOAD_DECLINE;
	}

	snprintf(new_global_config.data_directory, sizeof(new_global_config.data_directory) - 1, "%s/skype", ast_config_AST_SPOOL_DIR);

	for (v = ast_variable_browse(cfg, "general"); v; v = v->next) {
		if (!strcasecmp(v->name, "default_user")) {
			ast_copy_string(new_global_config.defaultuser, v->value, sizeof(new_global_config.defaultuser));
		} else if (!strcasecmp(v->name, "debug")) {
			new_global_config.debug = ast_true(v->value);
		} else if (!strcasecmp(v->name, "engine_directory")) {
			if (reload && strcasecmp(global_config.engine_directory, v->value)) {
				ast_log(LOG_WARNING, "Engine directory changes during a reload are not supported\n");
				ast_copy_string(new_global_config.engine_directory, global_config.engine_directory, sizeof(new_global_config.engine_directory));
			} else {
				ast_copy_string(new_global_config.engine_directory, v->value, sizeof(new_global_config.engine_directory));
			}
		} else if (!strcasecmp(v->name, "bind_address")) {
			if (reload && strcasecmp(global_config.bind_address, v->value)) {
				ast_log(LOG_WARNING, "Bind address changes during a reload are not supported\n");
				ast_copy_string(new_global_config.bind_address, global_config.bind_address, sizeof(new_global_config.bind_address));
			} else {
				ast_copy_string(new_global_config.bind_address, v->value, sizeof(new_global_config.bind_address));
			}
		} else if (!strcasecmp(v->name, "rtp_address")) {
			ast_copy_string(new_global_config.rtp_address, v->value, sizeof(new_global_config.rtp_address));
		} else if (!strcasecmp(v->name, "bind_port")) {
			ast_copy_string(new_global_config.bind_port, v->value, sizeof(new_global_config.bind_port));
		} else if (!strcasecmp(v->name, "disable_tcpauto")) {
			if (reload && (!global_config.disable_tcpauto != !ast_true(v->value))) {
				ast_log(LOG_WARNING, "disable_tcpauto changes during a reload are not supported\n");
				new_global_config.disable_tcpauto = global_config.disable_tcpauto;
			} else {
				new_global_config.disable_tcpauto = ast_true(v->value);
			}
		} else if (!strcasecmp(v->name, "disable_udp")) {
			if (reload && (!global_config.disable_udp != !ast_true(v->value))) {
				ast_log(LOG_WARNING, "disable_udp changes during a reload are not supported\n");
				new_global_config.disable_udp = global_config.disable_udp;
			} else {
				new_global_config.disable_udp = ast_true(v->value);
			}
		} else if (!strcasecmp(v->name, "https_proxy")) {
			ast_copy_string(new_global_config.https_proxy, v->value, sizeof(new_global_config.https_proxy));
		} else if (!strcasecmp(v->name, "https_proxy_user")) {
			ast_copy_string(new_global_config.https_proxy_user, v->value, sizeof(new_global_config.https_proxy_user));
		} else if (!strcasecmp(v->name, "https_proxy_password")) {
			ast_copy_string(new_global_config.https_proxy_password, v->value, sizeof(new_global_config.https_proxy_password));
		} else if (!strcasecmp(v->name, "socks5_proxy")) {
			ast_copy_string(new_global_config.socks5_proxy, v->value, sizeof(new_global_config.socks5_proxy));
		} else if (!strcasecmp(v->name, "socks5_proxy_user")) {
			ast_copy_string(new_global_config.socks5_proxy_user, v->value, sizeof(new_global_config.socks5_proxy_user));
		} else if (!strcasecmp(v->name, "socks5_proxy_password")) {
			ast_copy_string(new_global_config.socks5_proxy_password, v->value, sizeof(new_global_config.socks5_proxy_password));
		}
	}

	/* try to ensure that the data_directory is usable */
	if (stat(new_global_config.data_directory, &statbuf) == 0) {
		/* it exists, make sure it is a directory */
		if (!S_ISDIR(statbuf.st_mode)) {
			ast_log(LOG_ERROR, "Path '%s' is not a directory. Cannot continue.\n", new_global_config.data_directory);
			return AST_MODULE_LOAD_DECLINE;
		}
	} else {
		/* it doesn't exist, try to create it */
		if (ast_mkdir(new_global_config.data_directory, 0777)) {
			ast_log(LOG_ERROR, "Path '%s' does not exist and could not be created (%s). Cannot continue.\n",
				new_global_config.data_directory, strerror(errno));
		}
	}

	/* try to ensure that the engine_directory is usable */
	if (stat(new_global_config.engine_directory, &statbuf) == 0) {
		/* it exists, make sure it is a directory */
		if (!S_ISDIR(statbuf.st_mode)) {
			ast_log(LOG_ERROR, "Path '%s' specified for 'engine_directory' is not a directory. Cannot continue.\n", new_global_config.engine_directory);
			return AST_MODULE_LOAD_DECLINE;
		}
	} else {
		/* it doesn't exist, try to create it */
		if (ast_mkdir(new_global_config.engine_directory, 0777)) {
			ast_log(LOG_ERROR, "Path '%s' specified for 'engine_directory' does not exist and could not be created (%s). Cannot continue.\n",
				new_global_config.engine_directory, strerror(errno));
		}
	}

	/* attempt to convert bind_address string into an IP address */
	if (!(hp = ast_gethostbyname(new_global_config.bind_address, &ahp))) {
		ast_log(LOG_ERROR, "Invalid value '%s' specified for 'bind_address'. Cannot continue.\n", new_global_config.bind_address);
	} else {
		new_global_config.bind_sockaddr_in.sin_family = AF_INET;
		memcpy(&new_global_config.bind_sockaddr_in.sin_addr, hp->h_addr, sizeof(new_global_config.bind_sockaddr_in.sin_addr));
	}

	/* validate bind_port */
	if ((sscanf(new_global_config.bind_port, "%ud", &bind_port) != 1) || 
	    (bind_port > 65535)) {
		ast_log(LOG_ERROR, "Invalid value '%s' specified for 'bind_port'. Cannot continue.\n", new_global_config.bind_port);
	} else {
		new_global_config.bind_sockaddr_in.sin_port = htons(bind_port);
	}

	/* attempt to convert rtp_address string into an IP address */
	if (!(hp = ast_gethostbyname(new_global_config.rtp_address, &ahp))) {
		ast_log(LOG_ERROR, "Invalid value '%s' specified for 'rtp_address'. Cannot continue.\n", new_global_config.rtp_address);
	} else {
		new_global_config.rtp_sockaddr_in.sin_family = AF_INET;
		memcpy(&new_global_config.rtp_sockaddr_in.sin_addr, hp->h_addr, sizeof(new_global_config.rtp_sockaddr_in.sin_addr));
	}

	/* ensure https_proxy values seem sane */
	if (ast_strlen_zero(new_global_config.https_proxy)) {
		if (!ast_strlen_zero(new_global_config.https_proxy_user)) {
			ast_log(LOG_WARNING, "Username '%s' specified for 'https_proxy_user' without specifying 'https_proxy'. Ignored.\n", new_global_config.https_proxy_user);
			new_global_config.https_proxy_user[0] = '\0';
		}
		if (ast_strlen_zero(new_global_config.https_proxy_user)) {
			if (!ast_strlen_zero(new_global_config.https_proxy_password)) {
				ast_log(LOG_WARNING, "Password '%s' specified for 'https_proxy_password' without specifying 'https_proxy_user'. Ignored.\n", new_global_config.https_proxy_password);
				new_global_config.https_proxy_password[0] = '\0';
			}
		} else if (!ast_strlen_zero(new_global_config.https_proxy_password)) {
			ast_log(LOG_WARNING, "Password '%s' specified for 'https_proxy_password' without specifying 'https_proxy'. Ignored.\n", new_global_config.https_proxy_password);
			new_global_config.https_proxy_password[0] = '\0';
		}
	}

	/* ensure socks5_proxy values seem sane */
	if (ast_strlen_zero(new_global_config.socks5_proxy)) {
		if (!ast_strlen_zero(new_global_config.socks5_proxy_user)) {
			ast_log(LOG_WARNING, "Username '%s' specified for 'socks5_proxy_user' without specifying 'socks5_proxy'. Ignored.\n", new_global_config.socks5_proxy_user);
			new_global_config.socks5_proxy_user[0] = '\0';
		}
		if (ast_strlen_zero(new_global_config.socks5_proxy_user)) {
			if (!ast_strlen_zero(new_global_config.socks5_proxy_password)) {
				ast_log(LOG_WARNING, "Password '%s' specified for 'socks5_proxy_password' without specifying 'socks5_proxy_user'. Ignored.\n", new_global_config.socks5_proxy_password);
				new_global_config.socks5_proxy_password[0] = '\0';
			}
		} else if (!ast_strlen_zero(new_global_config.socks5_proxy_password)) {
			ast_log(LOG_WARNING, "Password '%s' specified for 'socks5_proxy_password' without specifying 'socks5_proxy'. Ignored.\n", new_global_config.socks5_proxy_password);
			new_global_config.socks5_proxy_password[0] = '\0';
		}
	}

	for (cat = ast_category_browse(cfg, NULL);
	     cat;
	     cat = ast_category_browse(cfg, cat)) {
		if (!strcasecmp(cat, "general")) {
			continue;
		}

		if ((user = build_user(cat, ast_variable_browse(cfg, cat), reload))) {
			ao2_ref(user, -1);
		}
	}

	if (reload) {
		ao2_callback(users, OBJ_UNLINK | OBJ_NODATA | OBJ_MULTIPLE, reload_cleanup, NULL);
	}

	global_config = new_global_config;

	sfa_set_global_debug(GLOBAL_DEBUG);

	return AST_MODULE_LOAD_SUCCESS;
}

static int set_sfa_directories(void)
{
	char license_path[PATH_MAX];

	if (sfa_set_data_directory(global_config.data_directory)) {
		ast_log(LOG_ERROR, "Unable to set 'data_directory' to '%s'.\n", global_config.data_directory);
		return -1;
	}

	if (sfa_set_engine_directory(global_config.engine_directory)) {
		ast_log(LOG_ERROR, "Unable to set 'engine_directory' to '%s'.\n", global_config.engine_directory);
		return -1;
	}

	ast_copy_string(license_path, ast_config_AST_VAR_DIR, sizeof(license_path));
	strncat(license_path, "/licenses", sizeof(license_path) - 1);
	if (sfa_set_license_directory(license_path)) {
		ast_log(LOG_ERROR, "Unable to set 'license_directory' to '%s'.\n", license_path);
		return -1;
	}

	return 0;
}


static int mark_deletepending(void *obj, void *arg, int flags)
{
	struct client_user_info *user = obj;
	
	user->deletepending = 1;
	
	return 0;
}

static void mark_all_users_deletepending(void)
{
	ao2_callback(users, 0, mark_deletepending, 0);
}

static void cleanup_users_deletepending(void)
{

	ao2_callback(users, OBJ_UNLINK | OBJ_NODATA | OBJ_MULTIPLE, cleanup_deletepending, 0);
}

static int reload(void)
{
	if (monitor_reload) {
		ast_log(LOG_WARNING, "Reload already in process\n");
		return -1;
	}

	sfa_on_reload();
	monitor_reload = 1;
	send_monitor_alert();

	return 0;
}

static void atexit_shutdown(void)
{
	/* Asterisk is exiting uncleanly, so we need to forcibly shutdown SFA. */
	sfa_shutdown();
}

static void shutdown_monitor_thread(void)
{
	monitor_shutdown = 1;
	pthread_cancel(monitor_thread);
	pthread_join(monitor_thread, NULL);
}

static int unload_module(void)
{
	ast_channel_unregister(&chan_tech);
	ast_rtp_glue_unregister(&skype_rtp_glue);
	ast_devstate_prov_del("Skype");
	ast_cli_unregister_multiple(cli, ARRAY_LEN(cli));
	ast_manager_unregister("SkypeBuddies");
	ast_manager_unregister("SkypeBuddy");
	ast_manager_unregister("SkypeAccountProperty");
	ast_manager_unregister("SkypeAddBuddy");
	ast_manager_unregister("SkypeRemoveBuddy");
	ast_manager_unregister("SkypeLicenseStatus");
	ast_manager_unregister("SkypeLicenseList");
	ast_manager_unregister("SkypeChatSend");
	ast_custom_function_unregister(&skype_callprop_function);
	ast_custom_function_unregister(&skype_accountprop_function);
	ast_custom_function_unregister(&skype_buddies_function);
	ast_custom_function_unregister(&skype_buddy_fetch_function);
	ast_custom_function_unregister(&skype_chat_receive_function);
	ast_unregister_application("SkypeChatSend");
	ast_unregister_application("SkypeTransfer");

	ao2_callback(chat_subscribers, OBJ_UNLINK | OBJ_NODATA | OBJ_MULTIPLE, NULL, NULL);
	ao2_ref(chat_subscribers, -1);
	chat_subscribers = NULL;
	
	if (!broken) {
		mark_all_users_deletepending();

		shutdown_monitor_thread();
		/* this function handles event processing for users
		 * by itself, so it must be run after the monitor thread
		 * is gone
		 */
		cleanup_users_deletepending();

		close(monitor_alert[0]);
		close(monitor_alert[1]);

		/* unref the *only* reference to the container of users, which will
		 * then unreference all the user objects in the container. presumably
		 * their reference counts will then reach zero, and they will be
		 * destroyed.
		 */
		users = unref_users(users);
		sfa_shutdown();
	}
	ast_unregister_atexit(atexit_shutdown);

	return 0;
}

static void control_socket_change(struct client_call_info *client_call, int control_socket)
{
	/* this will be called by sfa_call_process() when necessary, which means
	 * the appropriate locks on the client_call are already held
	 */
	client_call->owner->fds[FD_VE] = client_call->control_socket = control_socket;
}

static void dtmf_payload_type_selected(struct client_call_info *client_call, uint8_t type)
{
	/* this will be called by sfa_call_process() when necessary, which means
	 * the appropriate locks on the client_call are already held
	 */
	ast_rtp_codecs_payloads_set_m_type(ast_rtp_instance_get_codecs(client_call->rtp), client_call->rtp, type);
	ast_rtp_codecs_payloads_set_rtpmap_type(ast_rtp_instance_get_codecs(client_call->rtp), client_call->rtp, type, "audio", "telephone-event", 0);
}

static void audio_socket_address_selected(struct client_call_info *client_call, const struct sockaddr_in *address)
{
	/* this will be called by sfa_call_process() when necessary, which means
	 * the appropriate locks on the client_call are already held
	 */
	struct ast_sockaddr tmpaddr, us;
	struct sockaddr_in tmpus;

	ast_sockaddr_from_sin(&tmpaddr, address); 
	ast_rtp_instance_set_remote_address(client_call->rtp, &tmpaddr); 
	ast_rtp_instance_get_local_address(client_call->rtp, &us);
	ast_sockaddr_to_sin(&us, &tmpus);
	sfa_call_set_audio_socket_address(client_call->sfa_call, &tmpus);
}

static void codec_selected(struct client_call_info *client_call, const char *codec, unsigned int packet_size, unsigned int bit_rate)
{
	int new_format = 0;
	struct ast_codec_pref prefs;

	/* this will be called by sfa_call_process() when necessary, which means
	 * the appropriate locks on the client_call are already held
	 */
	if (!client_call->owner) {
		/* this should never ever happen */
		return;
	}

	if (!strcasecmp(codec, "PCMU")) {
		new_format = AST_FORMAT_ULAW;
	} else if (!strcasecmp(codec, "PCMA")) {
		new_format = AST_FORMAT_ALAW;
	} else if (!strcasecmp(codec, "G729")) {
		new_format = AST_FORMAT_G729A;
	}

	if (dynamic_codec_change) {
		prefs = ast_rtp_instance_get_codecs(client_call->rtp)->pref;
		ast_codec_pref_setsize(&prefs, new_format, packet_size);
		ast_rtp_codecs_packetization_set(ast_rtp_instance_get_codecs(client_call->rtp), client_call->rtp, &prefs);
	}

	if (client_call->owner->nativeformats != new_format) {
		client_call->owner->nativeformats = new_format;
		/* it would be better to actually change the readformat and writeformat here
		 * if possible
		 */
		ast_set_read_format(client_call->owner, client_call->owner->readformat);
		ast_set_write_format(client_call->owner, client_call->owner->writeformat);
	}

	if (!client_call->got_codec &&
	    (client_call->direction == DIRECTION_OUTGOING)) {
		if (client_call->delay_answer_until_codec &&
		    (client_call->owner->_state != AST_STATE_UP)) {
			/* an ANSWER was received for this call before the selected codec was known;
			 * now that the codec is known, forward the answer to Asterisk
			 */
				ast_queue_control(client_call->owner, AST_CONTROL_ANSWER);
		}
	} else if (!client_call->got_codec && 
		   (client_call->direction == DIRECTION_INCOMING)) {
#if 0
		/* now that the codec has been selected, a PBX can be started on the channel */
		if (ast_pbx_start(client_call->owner)) {
			ast_log(LOG_WARNING, "Unable to start PBX on channel '%s'\n", client_call->owner->name);
			ast_hangup(client_call->owner);
		}
#endif
	}

	client_call->got_codec = 1;
}

static enum ast_device_state buddystate2devstate(enum sfa_buddy_state buddy_state)
{
	switch (buddy_state) {
	case SFA_BUDDY_STATE_ONLINE:
	case SFA_BUDDY_STATE_SKYPE_ME:
	case SFA_BUDDY_STATE_OFFLINE_BUT_VM_ABLE:
	case SFA_BUDDY_STATE_OFFLINE_BUT_CF_ABLE:
	case SFA_BUDDY_STATE_AWAY:
		return AST_DEVICE_NOT_INUSE;
	case SFA_BUDDY_STATE_OFFLINE:
		return AST_DEVICE_UNAVAILABLE;
	case SFA_BUDDY_STATE_BLOCKED:
	case SFA_BUDDY_STATE_BLOCKED_SKYPEOUT:
	case SFA_BUDDY_STATE_NOT_AVAILABLE:
	case SFA_BUDDY_STATE_DO_NOT_DISTURB:
		return AST_DEVICE_BUSY;
	default:
		return AST_DEVICE_UNKNOWN;
	}
}

static enum ast_device_state devstate_provider(const char *data)
{
	char *buddyname = ast_strdupa(data ? data : "");
	char *username = strsep(&buddyname, "@");
	struct client_user_info *user;
	enum sfa_buddy_state buddy_state;

	if (ast_strlen_zero(buddyname) || ast_strlen_zero(username)) {
		ast_log(LOG_ERROR, "Skype device state requests must have both a user and a buddy specified. '%s' is not valid.\n", data);
		return AST_DEVICE_UNKNOWN;
	}

	if (!(user = find_active_user(username))) {
		ast_log(LOG_WARNING, "No such local Skype identity '%s' or user not active\n", username);
		return AST_DEVICE_UNKNOWN;
	}

	buddy_state = sfa_user_get_buddy_state(user->sfa_user, buddyname, user->autoadd_hint_contacts);

	ao2_ref(user, -1);

	return buddystate2devstate(buddy_state);
}

static void buddy_state_change(struct client_user_info *client_user, const char *buddy, enum sfa_buddy_state state)
{
	if (GLOBAL_DEBUG) {
		ast_log(LOG_DEBUG, "calling ast_device_state_changed() for user '%s' and buddy '%s' with state '%d'.\n", client_user->name, buddy, state);
	}
	ast_devstate_changed(buddystate2devstate(state), "Skype:%s@%s", client_user->name, buddy);
	manager_event(EVENT_FLAG_SYSTEM, "SkypeBuddyStatus",
		      "Buddy: Skype/%s@%s\r\n"
		      "BuddyStatus: %s\r\n", client_user->name, buddy, sfa_buddy_state2str(state));
}

static void account_status_change(struct client_user_info *client_user, enum sfa_account_status status)
{
	if (GLOBAL_DEBUG) {
		ast_log(LOG_DEBUG, "User '%s' status changed to %d\n", client_user->name, status);
	}

	if (!client_user->sfa_user) {
		ast_log(LOG_WARNING, "%s was unable to connect to the Skype network\n", client_user->name);
		return;
	}

	if (status == SFA_ACCOUNT_STATUS_LOGGED_IN) {
		char buf[100];
		struct ast_tm tm;
		struct timeval t = ast_tvnow();

		if (ast_db_get("Skype/Availability", client_user->name, buf, sizeof(buf))) {
			sfa_account_set_property(client_user->sfa_user, SFA_AVAILABILITY, buf, 0);
		}
		sfa_account_set_property(client_user->sfa_user, SFA_FULLNAME, client_user->fullname, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_LANGUAGES, client_user->language, 0); 
		sfa_account_set_property(client_user->sfa_user, SFA_COUNTRY, client_user->country, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_PROVINCE, client_user->province, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_CITY, client_user->city, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_PHONE_HOME, client_user->phone_home, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_PHONE_OFFICE, client_user->phone_office, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_PHONE_MOBILE, client_user->phone_mobile, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_EMAILS, client_user->email, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_HOMEPAGE, client_user->homepage, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_ABOUT, client_user->about, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_BIRTHDAY, client_user->birthday, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_MOOD_TEXT, client_user->mood_text, 0);
		sfa_account_set_property(client_user->sfa_user, SFA_WEBPRESENCE_POLICY, ast_true(client_user->webpresence) ? "1" : "0", 0);

		if (!ast_strlen_zero(client_user->gender)) {
			switch (client_user->gender[0]) {
			case 'M':
			case 'm':
				sfa_account_set_property(client_user->sfa_user, SFA_GENDER, "1", 0);
				break;
			case 'F':
			case 'f':
				sfa_account_set_property(client_user->sfa_user, SFA_GENDER, "2", 0);
				break;
			default:
				break;
			}
		}	

		if (!ast_strlen_zero(client_user->avatar_filename)) {
			int fd;
			char buffer[MAX_AVATAR_FILESIZE];
			size_t len;

			if (!(fd = open(client_user->avatar_filename, O_RDONLY))) {
				ast_log(LOG_ERROR, "Avatar file '%s' not found for %s\n", client_user->avatar_filename, client_user->name);
			} else {
				len = read(fd, buffer, MAX_AVATAR_FILESIZE);
				close(fd);
				sfa_account_set_property(client_user->sfa_user, SFA_AVATAR_IMAGE, buffer, len);
			}
		}

		ast_localtime(&t, &tm, ast_strlen_zero(client_user->timezone) ? NULL : client_user->timezone);
		snprintf(buf, sizeof(buf), "%ld", 24 * 3600 + tm.tm_gmtoff);
		sfa_account_set_property(client_user->sfa_user, SFA_TIMEZONE, buf, 0);

		sfa_account_commit_properties(client_user->sfa_user);
	}

	if (status == SFA_ACCOUNT_STATUS_LOGGED_IN || status == SFA_ACCOUNT_STATUS_LOGGED_OUT) {
		manager_event(EVENT_FLAG_SYSTEM, "SkypeAccountStatus",
			      "Username: %s\r\n"
			      "Status: %s\r\n", client_user->name, sfa_user_account_status(client_user->sfa_user));
	}

	if (status == SFA_ACCOUNT_STATUS_LOGGED_OUT) {
		enum sfa_account_logoutreason reason = sfa_user_account_logoutreason(client_user->sfa_user);
		ast_log(LOG_DEBUG, "%s logged out. LOGOUT REASON = %s\n", client_user->name, sfa_account_logoutreason2str(reason));
		if (reason == SFA_ACCOUNT_LOGOUTREASON_INCORRECT_PASSWORD) {
			ast_log(LOG_WARNING,
				"Account %s logged out due to an incorrect password or invalid Skype BCP account\n", client_user->name); 
			ast_verb(1, "The account could not log in due to an incorrect password. This could\n");
			ast_verb(1, "be due to the password being incorrect or because the account was\n");
			ast_verb(1, "not created by the Skype Business Control Panel by clicking 'Create business\n");
			ast_verb(1, "account'. Please be advised that the account that administers the BCP will not\n");
			ast_verb(1, "work, nor will normal accounts that have been invited by the BCP. To verify\n");
			ast_verb(1, "that the account is a BCP account, log into the Skype website at\n");
			ast_verb(1, "https://secure.skype.com/account/login with the account and verify that\n");
			ast_verb(1, "there is a box at the top of the screen saying 'You are a member of a\n");
			ast_verb(1, "Business Control Panel'. If that message isn't there, you will need to create\n");
			ast_verb(1, "a user from the Business Control Panel by clicking 'Add Members' and then\n");
			ast_verb(1, "'Create a business account'.\n");
		} else  if (reason == SFA_ACCOUNT_LOGOUTREASON_PASSWORD_HAS_CHANGED) {
			sfa_user_login(client_user->sfa_user, client_user->secret);
		}
	}
}

static int skype_transfer_exec(struct ast_channel *chan, const char *data)
{
	char *parse;
	struct ast_module_user *u;
	struct client_call_info *client_call = chan->tech_pvt;
	int res;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(to);
		AST_APP_ARG(msg);
	);

	u = ast_module_user_add(chan);
	
	if (ast_strlen_zero((const char *) data) || !(parse = ast_strdupa((const char *) data))) {
		ast_module_user_remove(u);
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.to)) {
		ast_module_user_remove(u);
		return -1;
	}
	
	if (client_call->owner->_state != AST_STATE_UP) {
		ast_answer(client_call->owner);
	}

	if ((res = sfa_transfer_call(client_call->sfa_call, args.to, args.msg))) {
		ast_log(LOG_WARNING, "Transfer to '%s' failed\n", args.to);
	}

	if (client_call->owner) {
		enum ast_control_transfer message = res ? AST_TRANSFER_FAILED : AST_TRANSFER_SUCCESS;
		ast_queue_control_data(client_call->owner, AST_CONTROL_TRANSFER, &message, sizeof(message));
	}

	ast_module_user_remove(u);
	return res;
}

static int skype_sendchat_exec(struct ast_channel *chan, const char *data)
{
	char *parse;
	struct ast_module_user *u;
	struct client_user_info *user;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(from);
		AST_APP_ARG(to);
		AST_APP_ARG(msg);
	);

	u = ast_module_user_add(chan);
	
	if (ast_strlen_zero(data) || !(parse = ast_strdupa(data))) {
		ast_module_user_remove(u);
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.from) || ast_strlen_zero(args.to) || ast_strlen_zero(args.msg)) {
		ast_module_user_remove(u);
		return -1;
	}

	if (!(user = find_active_user(args.from))) {
		ast_module_user_remove(u);
		return -1;
	}

	sfa_send_chat_message(user->sfa_user, args.to, args.msg);

	ast_module_user_remove(u);
	return 0;
}

static const char manshowbuddies_usage[] =
"Description: List a Skype user's buddy list.\n"
"Variables: \n"
"  ActionID: <id>  Action ID for this transaction. Will be returned.\n";

static const char manshowbuddy_usage[] =
"Description: Set a poperty on a Skype account.\n"
"Variables: \n"
"  ActionID: <id>    Action ID for this transaction. Will be returned.\n"
"  User    : <name>  Skype account name\n"
"  Variable: <property>=<value>\n";

static const char manaccountprop_usage[] =
"Description: Display all buddy properties that have been set.\n"
"Variables: \n"
"  ActionID: <id>    Action ID for this transaction. Will be returned.\n"
"  User    : <name>  Skype account name\n"
"  Buddy   : <name>  Name of buddy\n";

static const char manaddbuddy_usage[] =
"Description: Add a skype buddy to a user's buddy list"
"Variables: \n"
"  ActionID: <id>    Action ID for this transaction. Will be returned.\n"
"  User    : <name>  Skype account name\n"
"  Buddy   : <name>  Name of buddy\n"
"  AuthMsg : <msg>   Authentication request message\n";

static const char manremovebuddy_usage[] =
"Description: Remove a skype buddy from a user's buddy list"
"Variables: \n"
"  ActionID: <id>    Action ID for this transaction. Will be returned.\n"
"  User    : <name>  Skype account name\n"
"  Buddy   : <name>  Name of buddy\n";

static const char manlicenselist_usage[] =
"Description: Reports the total number of licensed channels\n"
"Variables: \n"
"  ActionID: <id>    Action ID for this transaction. Will be returned.\n";

static const char manlicensestatus_usage[] =
"Description: Lists the Skype for Asterisk licenses found and their status.\n"
"Variables: \n"
"  ActionID: <id>    Action ID for this transaction. Will be returned.\n";

static const char manskypechatsend_usage[] =
"Description: Send a Skype Chat message.\n"
"Variables: \n"
"  ActionID  : <id>    Action ID for this transaction. Will be returned.\n"
"  User      : <name>  Local skype account name\n"
"  Skypename : <name>  Remote skype account name\n"
"  Message   : <text>  Message text\n";

static const char *app_sendchat_synopsis = "SkypeChatSend(from,to,message)";
static const char *app_sendchat_descrip =
"SkypeChatSend(from,to,message)\n"
"  from - Skype account to send the message from\n"
"  to - Skype account to send the message to\n"
"  message - The message to send\n";

static const char *app_skypetransfer_synopsis = "SkypeTransfer(to,topic)";
static const char *app_skypetransfer_descrip =
"SkypeTransfer(to,topic)\n"
"  to - Skype account to transfer this call to\n"
"  topic - The topic of the transferred call\n";

static enum ast_module_load_result load_module(void)
{
	enum ast_module_load_result config_result;
	struct sfa_callbacks cb = {
		.logger = ast_log,
		.new_call = new_call,
		.queue_event = queue_event,
		.authorization_request = authrequest,
		.control_socket_change = control_socket_change,
		.codec_selected = codec_selected,
		.buddy_state_change = buddy_state_change,
		.account_status_change = account_status_change,
		.audio_socket_address_selected = audio_socket_address_selected,
		.dtmf_payload_type_selected = dtmf_payload_type_selected,
		.new_chat_message = new_chat_message,
	};
	int res = AST_MODULE_LOAD_SUCCESS;

	/* recent versions of Asterisk have bug fixes that allow the packet size
	 * of RTP streams to be safely changed on the fly; to know whether that
	 * support is available or not, we'll check at run time for a smoother
	 * API call that was added as part of the fix. if the support is not
	 * available, we won't try to adjust the packet size when requested by
	 * Skype.
	 */

	if (dlsym(RTLD_DEFAULT, "ast_smoother_reconfigure")) {
		dynamic_codec_change = 1;
	}

	if (!(chat_subscribers = ao2_container_alloc(13, chat_subscriber_hash_fn, chat_subscriber_cmp_fn))) {
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_codec_pref_append(&default_prefs, AST_FORMAT_ULAW);

	if (!(users = ao2_container_alloc(USER_BUCKETS, user_hash_cb, user_cmp_cb))) {
		return AST_MODULE_LOAD_DECLINE;
	}

	if (sfa_set_callbacks(&cb)) {
		ast_log(LOG_ERROR, "Unable to set callback functions in Skype For Asterisk library.\n");
		users = unref_users(users);
		return AST_MODULE_LOAD_DECLINE;
	}

	if ((config_result = load_config(0)) != AST_MODULE_LOAD_SUCCESS) {
		return config_result;
	}

	if (set_sfa_directories()) {
		sfa_shutdown();
		users = unref_users(users);
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_register_atexit(atexit_shutdown);

	if (sfa_startup(SFA_STARTUP_START_SKYHOST)) {
		ast_log(LOG_ERROR, "Unable to start Skype For Asterisk library.\n");
		users = unref_users(users);
		res = AST_MODULE_LOAD_DECLINE;
	} else if (set_sfa_globals()) {
		sfa_shutdown();
		users = unref_users(users);
		res = AST_MODULE_LOAD_DECLINE;
	} else if (pipe(monitor_alert)) {
		ast_log(LOG_ERROR, "Unable to create monitor alert pipe (%s).\n", strerror(errno));
		sfa_shutdown();
		users = unref_users(users);
		res = AST_MODULE_LOAD_DECLINE;
	} else if (ast_pthread_create_background(&monitor_thread, NULL, do_monitor, NULL) < 0) {
		ast_log(LOG_ERROR, "Unable to start monitor thread (%s).\n", strerror(errno));
		close(monitor_alert[0]);
		close(monitor_alert[1]);
		sfa_shutdown();
		users = unref_users(users);
		res = AST_MODULE_LOAD_DECLINE;
	} else if ((ast_rtp_glue_register(&skype_rtp_glue))) {
		ast_log(LOG_ERROR, "Unable to register RTP protocol structure.\n");
		shutdown_monitor_thread();
		close(monitor_alert[0]);
		close(monitor_alert[1]);
		sfa_shutdown();
		users = unref_users(users);
		res = AST_MODULE_LOAD_DECLINE;
	}

	if (res == AST_MODULE_LOAD_SUCCESS) {
		connect_users();
		send_monitor_alert();
		ast_channel_register(&chan_tech);
		ast_devstate_prov_add("Skype", devstate_provider);
	} else {
		broken = 1;
	}

	/* Even if we had errors, go ahead and load CLI/Manager functions so that license
	 * errors can be queried */
	ast_cli_register_multiple(cli, ARRAY_LEN(cli));
	ast_manager_register2("SkypeLicenseStatus", EVENT_FLAG_CALL, license_status_manager,
		"Skype for Asterisk license status", manlicensestatus_usage);
	ast_manager_register2("SkypeLicenseList", EVENT_FLAG_CALL, license_list_manager,
		"Skype for Asterisk license list", manlicenselist_usage);
	ast_manager_register2("SkypeChatSend", EVENT_FLAG_SYSTEM, send_chat_manager,
		"Send a Skype Chat message", manskypechatsend_usage);
	
	if (!broken) {
		ast_manager_register2("SkypeBuddies", EVENT_FLAG_SYSTEM, show_buddies_manager,
			"List Skype buddies", manshowbuddies_usage);
		ast_manager_register2("SkypeBuddy", EVENT_FLAG_SYSTEM, show_buddy_manager,
			"Set a Skype account property", manaccountprop_usage);
		ast_manager_register2("SkypeAccountProperty", EVENT_FLAG_SYSTEM,
			account_property_write_manager, "Display Skype Buddy properties", manshowbuddy_usage);
		ast_manager_register2("SkypeAddBuddy", EVENT_FLAG_SYSTEM, add_buddy_manager,
			"Add a skype buddy", manaddbuddy_usage);
		ast_manager_register2("SkypeRemoveBuddy", EVENT_FLAG_SYSTEM, remove_buddy_manager,
			"Remove a skype buddy", manremovebuddy_usage);

		ast_custom_function_register(&skype_callprop_function);
		ast_custom_function_register(&skype_accountprop_function);
		ast_custom_function_register(&skype_buddies_function);
		ast_custom_function_register(&skype_buddy_fetch_function);
		ast_custom_function_register(&skype_chat_receive_function);

		ast_register_application("SkypeChatSend", skype_sendchat_exec, app_sendchat_synopsis, app_sendchat_descrip);
		ast_register_application("SkypeTransfer", skype_transfer_exec, app_skypetransfer_synopsis, app_skypetransfer_descrip);

	}


	return AST_MODULE_LOAD_SUCCESS;
}

static unsigned int monitor_array_used = 0;
static struct pollfd *monitor_pfds = NULL;
static struct client_user_info **monitor_users = NULL;

static int add_user_to_monitor_arrays(void *obj, void *arg, int flags)
{
		struct client_user_info *user = obj;

		if (!user->sfa_user || user->broken) {
			return 0;
		}

		monitor_users[monitor_array_used] = user;
		ao2_ref(user, +1);
		monitor_pfds[monitor_array_used].fd = user->sfa_socket;
		monitor_array_used++;

		return 0;
}

static void *do_monitor(void *data)
{
	unsigned int array_size = 0;
	unsigned int rebuild_needed = 1;
	unsigned int x;
	int res;

	if (GLOBAL_DEBUG) {
		ast_log(LOG_DEBUG, "thread starting\n");
	}

	for (;;) {
		if (monitor_reload) {
			ao2_lock(users);
			load_config(1);
			ao2_unlock(users);

			monitor_reload = 0;
			rebuild_needed = 1;
		}

		if (rebuild_needed) {
			/* array needs to hold the user sockets plus the alertpipe */
			unsigned int array_needed = ao2_container_count(users) + 1;

			if (GLOBAL_DEBUG) {
				ast_log(LOG_DEBUG, "rebuilding poll arrays\n");
			}

			/* check if the existing arrays are large enough, if not,
			 * reallocate
			 */
			if (array_needed > array_size) {
				if (array_size) {
					for (x = 0; x < monitor_array_used; x++) {
						ao2_ref(monitor_users[x], -1);
					}
					ast_free(monitor_users);
					ast_free(monitor_pfds);
					array_size = 0;
				}

				/* make the array a bit larger than needed */
				array_size = (array_needed + 128);

				if (GLOBAL_DEBUG) {
					ast_log(LOG_DEBUG, "allocating arrays for %d elements\n", array_size);
				}

				if (!(monitor_pfds = ast_calloc(sizeof(*monitor_pfds), array_size))) {
					/* PUNT */
				}
				if (!(monitor_users = ast_calloc(sizeof(*monitor_users), array_size))) {
					/* PUNT */
				}
			}

			monitor_array_used = 0;
			monitor_pfds[monitor_array_used++].fd = monitor_alert[0];

			ao2_callback(users, 0, add_user_to_monitor_arrays, NULL);

			rebuild_needed = 0;
		}

		/* we can reuse the existing pollfd array, but we need to clear
		 * the revents fields first
		 */
		for (x = 0; x < monitor_array_used; x++) {
			monitor_pfds[x].events = POLLIN;
			monitor_pfds[x].revents = 0;
		}

		if (GLOBAL_DEBUG) {
			ast_log(LOG_DEBUG, "entering poll for %d fds\n", monitor_array_used);
		}

		/* now we can actually poll on all the fds, and we can wait forever
		 * since the only actions this thread can take will be triggered by
		 * activity on one or more fds
		 */
		res = poll(monitor_pfds, monitor_array_used, -1);

		if (GLOBAL_DEBUG) {
			ast_log(LOG_DEBUG, "poll returned %d\n", res);
		}

		/* check if we should shutdown */
		if (monitor_shutdown) {
			for (x = 0; x < monitor_array_used; x++) {
				ao2_ref(monitor_users[x], -1);
			}
			ast_free(monitor_pfds);
			ast_free(monitor_users);
			break;
		}
		
		/* check if we should reload the configuration */
		if (monitor_reload) {
			ast_log(LOG_NOTICE, "Starting reload\n");
			continue;
		}

		/* if an error occurred, it's best to just loop around and rebuild
		 * the arrays, because it's likely that a socket we were waiting on
		 * was closed or something else important occurred
		 */
		if (res < 0) {
			rebuild_needed = 1;
			ast_log(LOG_WARNING, "got error from poll(), rebuilding: %s", strerror(errno));
			continue;
		}

		/* if we got here, at least one fd had activity on it, but we
		 * want to check the alertpipe first, because if we got an alert
		 * we may very well end up not including some of the users
		 * we had before
		 */
		if (monitor_pfds[0].revents & POLLIN) {
			char buf[1024];

			rebuild_needed = 1;

			if (GLOBAL_DEBUG) {
				ast_log(LOG_DEBUG, "got alert, rebuilding\n");
			}

			if (read(monitor_pfds[0].fd, buf, sizeof(buf)) != sizeof(buf)) {
				/* we know this won't actually happen unless the system
				 * is collapsing...
				 */
			}
			continue;
		}

		/* finally, process the fds that had data available */
		for (x = 1; res && (x < monitor_array_used); x++) {
			if (!monitor_pfds[x].revents) {
				continue;
			}

			res--;

			/* if there is an error of some kind on the socket,
			 * rebuild our arrays because that channel is probably
			 * marked for removal
			 */

			if (monitor_pfds[x].revents & (POLLERR | POLLHUP | POLLNVAL)) {
				rebuild_needed = 1;
				ast_log(LOG_WARNING, "fd %d had an error, marking broken and queuing rebuild\n", monitor_pfds[x].fd);
				monitor_users[x]->broken = 1;
				continue;
			}

			if (monitor_users[x]->sfa_user) {
				sfa_user_run(monitor_users[x]->sfa_user, SFA_USER_RUN_NONBLOCK);
			}
		}
	}

	return NULL;
}

void unlink_client_call(struct client_call_info *pvt)
{
	if (GLOBAL_DEBUG) {
		ast_log(LOG_DEBUG, "client_call: %p\n", pvt);
	}

	if (pvt->sfa_call) {
		unset_client_call(pvt);
	}
	ao2_unlink(pvt->user->calls, pvt);
}

void destroy_client_call(void *obj)
{
	struct client_call_info *pvt = obj;

	if (GLOBAL_DEBUG) {
		ast_log(LOG_DEBUG, "channel %p\n", pvt);
	}

	if (pvt->sfa_call) {
		sfa_call_hangup(pvt->sfa_call);
	}
	ao2_ref(pvt->user, -1);
	if (pvt->cid_name) {
		ast_free(pvt->cid_name);
	}
	if (pvt->cid_num) {
		ast_free(pvt->cid_num);
	}
	if (pvt->control_socket > -1) {
		close(pvt->control_socket);
	}
	if (pvt->rtp) {
		ast_rtp_instance_destroy(pvt->rtp);
	}
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Skype For Asterisk Channel Driver",
	.load = load_module, 
	.unload = unload_module, 
	.reload = reload,
);
