#ifndef _SKYPEFORASTERISK_H
#define _SKYPEFORASTERISK_H

#define attr_const __attribute__((const))
#define attr_pure __attribute__((pure))
#define attr_malloc __attribute__((malloc))

#include <netinet/in.h>

#if defined(__cplusplus) || defined(c_plusplus)
extern "C" {
#endif

#define MAX_AVATAR_FILESIZE 8192

#define EXT_API __attribute__((externally_visible))

struct sfa_user_info;
struct sfa_call_info;
struct client_user_info;
struct client_call_info;

/*!
 * \brief Send 'ringing' indication on an incoming call.
 *
 * \param sfa_call Pointer to the call
 * \return nothing
 */
void EXT_API sfa_call_ringback(struct sfa_call_info *sfa_call);

/*!
 * \brief Answer an incoming call.
 *
 * \param sfa_call Pointer to the call
 * \return nothing
 */
void EXT_API sfa_call_answer(struct sfa_call_info *sfa_call);

/*!
 * \brief Send a DTMF digit to the other party in a call.
 *
 * \param sfa_call Pointer to the call
 * \param digit The digit to be sent, as an ASCII character.
 * \return nothing
 *
 * \note Only digits '0' through '9', '*' and '#' are supported.
 */
void EXT_API sfa_call_send_dtmf(struct sfa_call_info *sfa_call, char digit);

/*!
 * \brief Associate a 'client data' pointer with a call.
 *
 * \param sfa_call Pointer to the call
 * \param client_call Client's data pointer
 * \return nothing
 *
 * \note The client_call pointer will be used by callbacks that provide events
 *       related to the call, so the client can identify the proper call.
 */
void EXT_API sfa_call_set_client_call(struct sfa_call_info *sfa_call, struct client_call_info *client_call);

/*!
 * \brief Ring (call) another party.
 *
 * \param sfa_call Pointer to the call
 * \param ident Identity of the party to be called
 * \param media_socket File descriptor (handle) of the socket associated with this call (output).
 * \return 0 on success, -1 if call cannot be placed, -2 if 'ident' refers to a disallowed destination
 *
 */
int EXT_API sfa_call_ring(struct sfa_call_info *sfa_call, const char *ident, int *media_socket, int *chanalert);

/*!
 * \brief Hangup a call.
 *
 * \param sfa_call Pointer to the call
 * \return nothing
 *
 * \note When this function returns, the sfa_call pointer is no longer valid and must not
 *       be dereferenced.
 */
void EXT_API sfa_call_hangup(struct sfa_call_info *sfa_call);

/*!
 * \brief Create a new call structure.
 *
 * \param sfa_user The user who will own (originate) the call.
 * \return pointer to the call structure on success, NULL on failure
 *
 */
struct sfa_call_info * EXT_API sfa_call_new(struct sfa_user_info *sfa_user) attr_malloc;

/*!
 * \brief Check whether a call is active or inactive.
 *
 * \param sfa_call Pointer to the call
 * \return non-zero if active, 0 if inactive
 */
int EXT_API sfa_call_active(const struct sfa_call_info *sfa_call) attr_const;

/*!
 * \brief Place a call on hold.
 *
 * \param sfa_call Pointer to the call
 * \return nothing
 */
void EXT_API sfa_call_hold(struct sfa_call_info *sfa_call);

/*!
 * \brief Retrieve a call from hold.
 *
 * \param sfa_call Pointer to the call
 * \return nothing
 */
void EXT_API sfa_call_unhold(struct sfa_call_info *sfa_call);

/*!
 * \brief Process control information from the control socket for a call.
 *
 * \param sfa_call Pointer to the call
 * \retval -1 on error
 * \retval 0 on success
 *
 * The client should call this function whenever select() or poll() indicate that the
 * control_socket for the call is 'ready' for reading.
 *
 * \note During control information processing, this function may call the client's
 *       control_socket_change callback function; if this occurs, the client must not attempt
 *       to use the previous control socket descriptor any longer, but must switch to the new
 *       one for all future select() or poll() operations.
 */
int EXT_API sfa_call_process(struct sfa_call_info *sfa_call);

/*!
 * \brief Process channel events while ast_read holds the channel lock
 * param sfa_call Pointer to the call
 * \retval -1 on error
 * \retval 0 on success
 */
int EXT_API sfa_channel_process(struct sfa_call_info *sfa_call);

/*!
 * \brief Accounts and Contacts share a lot of the same properties. To avoid
 * defining them twice, just define them all here and differntiate in the
 * specific get_property and set_property functions. Also, contacts are
 * frequently queried via a CallMember, so it is easier just to lump all of
 * the properties into a single enum.
 */
enum sfa_property {
	SFA_NO_SUCH_PROPERTY = 0,
	SFA_CALL_TOPIC,          /*! call topic */
	SFA_CALL_TOKEN,              /*! access token */
	SFA_SKYPENAME,
	SFA_PSTNNUMBER,
	SFA_FULLNAME,
	SFA_BIRTHDAY,
	SFA_GENDER,
	SFA_LANGUAGES,
	SFA_COUNTRY,
	SFA_PROVINCE,
	SFA_CITY,
	SFA_PHONE_HOME,
	SFA_PHONE_OFFICE,
	SFA_PHONE_MOBILE,
	SFA_EMAILS,
	SFA_HOMEPAGE,
	SFA_ABOUT,
	SFA_AVATAR_IMAGE,
	SFA_STATUS,
	SFA_AVAILABILITY,
	SFA_TIMEZONE,
	SFA_SKYPEOUT_BALANCE_CURRENCY,
	SFA_SKYPEOUT_BALANCE,
	SFA_FORWARDED_BY,
	SFA_TARGET_IDENTITY,
	SFA_TRANSFERRED_BY,
	SFA_TRANSFERRED_TO,
	SFA_MOOD_TEXT,
	SFA_LOGOUTREASON,
	SFA_WEBPRESENCE_POLICY,
/* currently unused
	SFA_RICH_MOOD_TEXT,
	SFA_CAPABILITIES,
	SFA_PWDGCHANGESTATUS,
	SFA_COMMITSTATUS,
	SFA_SUGGESTED_SKYPENAME,
	SFA_SKYPEIN_NUMBERS,
	SFA_CBLSYNCSTATUS,
	SFA_SERVICE_PROVIDER_INFO,
	SFA_REGISTRATION_TIMESTAMP,
	SFA_NR_OF_OTHER_INSTANCES,
	SFA_OFFLINE_CALLFORWARD,
	SFA_CHAT_POLICY,
	SFA_CALL_POLICY,
	SFA_PSTN_CALL_POLICY,
	SFA_AVATAR_POLICY,
	SFA_BUDDYCOUNT_POLICY,
	SFA_TIMEZONE_POLICY,
	SFA_PHONENUMBERS_POLICY,
	SFA_VOICEMAIL_POLICY,
*/
};

enum sfa_buddy_state {
	SFA_BUDDY_STATE_UNKNOWN,
	SFA_BUDDY_STATE_PENDINGAUTH,
	SFA_BUDDY_STATE_BLOCKED,
	SFA_BUDDY_STATE_BLOCKED_SKYPEOUT,
	SFA_BUDDY_STATE_SKYPEOUT,
	SFA_BUDDY_STATE_OFFLINE,
	SFA_BUDDY_STATE_ONLINE,
	SFA_BUDDY_STATE_AWAY,
	SFA_BUDDY_STATE_NOT_AVAILABLE,
	SFA_BUDDY_STATE_DO_NOT_DISTURB,
	SFA_BUDDY_STATE_SKYPE_ME,
	SFA_BUDDY_STATE_OFFLINE_BUT_VM_ABLE,
	SFA_BUDDY_STATE_OFFLINE_BUT_CF_ABLE,
	SFA_BUDDY_STATE_CONNECTING, /* Account availability only */
	SFA_BUDDY_STATE_INVISIBLE,	/* Account availability only */
};

enum sfa_account_logoutreason {
	SFA_ACCOUNT_LOGOUTREASON_UNKNOWN,
	SFA_ACCOUNT_LOGOUTREASON_LOGOUT_CALLED,                  /**< Manual logout (or unknown reason, from previous session) */
	SFA_ACCOUNT_LOGOUTREASON_HTTPS_PROXY_AUTH_FAILED,            /**< Synchronous error at login/registration */
	SFA_ACCOUNT_LOGOUTREASON_SOCKS_PROXY_AUTH_FAILED,            /**< Synchronous error at login/registration */
	SFA_ACCOUNT_LOGOUTREASON_P2P_CONNECT_FAILED,                 /**< Synchronous error at login/registration */
	SFA_ACCOUNT_LOGOUTREASON_SERVER_CONNECT_FAILED,              /**< Synchronous error at login/registration */
	SFA_ACCOUNT_LOGOUTREASON_SERVER_OVERLOADED,                  /**< Synchronous error at login/registration */
	SFA_ACCOUNT_LOGOUTREASON_DB_IN_USE,
	SFA_ACCOUNT_LOGOUTREASON_INVALID_SKYPENAME,                  /**< Synchronous error at registration */
	SFA_ACCOUNT_LOGOUTREASON_INVALID_EMAIL,                      /**< Synchronous error at registration */
	SFA_ACCOUNT_LOGOUTREASON_UNACCEPTABLE_PASSWORD,
	SFA_ACCOUNT_LOGOUTREASON_SKYPENAME_TAKEN,                    /**< Synchronous error at registration */
	SFA_ACCOUNT_LOGOUTREASON_REJECTED_AS_UNDERAGE,               /**< Synchronous error at registration */
	SFA_ACCOUNT_LOGOUTREASON_NO_SUCH_IDENTITY,                   /**< Synchronous error at login.
							In practice this error is never returned, when a user name that doesn't exist 
							is used, INCORRECT_PASSWORD is returned. */
	SFA_ACCOUNT_LOGOUTREASON_INCORRECT_PASSWORD,                 /**< Synchronous error at login */
	SFA_ACCOUNT_LOGOUTREASON_TOO_MANY_LOGIN_ATTEMPTS,            /**< Synchronous error at login */
	SFA_ACCOUNT_LOGOUTREASON_PASSWORD_HAS_CHANGED,               /**< Asynchronous errors (can happen anytime while logged in). */
	SFA_ACCOUNT_LOGOUTREASON_PERIODIC_UIC_UPDATE_FAILED,         /**< Asynchronous errors (can happen anytime while logged in).
							UIC stands for "user identity certificate". The way skype "login" works is that server
							issues the UIC to the user, which is like a "passport" that could be used to prove your
							identity to others. The user may choose to save that UIC locally (by setting savePwd=true),
							in which case it needs to be periodically re-issued (because it has limited validity).
							If that periodic update fails (because of server issues) and the "best before" of
							the UIC runs out, SkyHost gives this error. Try LoginWithPassword again. */
	SFA_ACCOUNT_LOGOUTREASON_DB_DISK_FULL,
	SFA_ACCOUNT_LOGOUTREASON_DB_IO_ERROR,
	SFA_ACCOUNT_LOGOUTREASON_DB_CORRUPT,
	SFA_ACCOUNT_LOGOUTREASON_DB_FAILURE,                         /**< Asynchronous errors (can happen anytime while logged in).
							Access to the database files in .Skype directory has failed. */
	SFA_ACCOUNT_LOGOUTREASON_INVALID_APP_ID,
	SFA_ACCOUNT_LOGOUTREASON_APP_ID_BLACKLISTED
};

int EXT_API sfa_call_get_property(struct sfa_call_info *sfa_call, enum sfa_property property, char *buffer, size_t buffer_size);
void EXT_API sfa_call_set_property(struct sfa_call_info *sfa_call, enum sfa_property property, const char *value);
int EXT_API sfa_account_get_property(struct sfa_user_info *user, enum sfa_property property, char *buffer, size_t buffer_size);
void EXT_API sfa_account_set_property(struct sfa_user_info *user, enum sfa_property property, const char *value, size_t len);
void EXT_API sfa_account_commit_properties(struct sfa_user_info *user);
enum sfa_property EXT_API sfa_str2property(const char *str);

void EXT_API sfa_call_set_audio_socket_address(struct sfa_call_info *sfa_call, const struct sockaddr_in *address);

struct sfa_user_info * EXT_API sfa_user_new(const char *name, struct client_user_info *client_user,
					    unsigned int debug, unsigned int vedebug) attr_malloc;
void EXT_API sfa_user_login(struct sfa_user_info *sfa_user, const char *secret);
void EXT_API sfa_user_logout(struct sfa_user_info *sfa_user);
void EXT_API sfa_user_shutdown(struct sfa_user_info *sfa_user);
void EXT_API sfa_user_destroy(struct sfa_user_info *sfa_user);

int EXT_API sfa_user_set_codecs(struct sfa_user_info *sfa_user, const char *codecs);

enum sfa_user_run_mode {
	SFA_USER_RUN_BLOCK,
	SFA_USER_RUN_NONBLOCK,
};

void EXT_API sfa_user_run(struct sfa_user_info *sfa_user, enum sfa_user_run_mode run_mode);
int EXT_API sfa_user_get_socket(const struct sfa_user_info *sfa_user) attr_const;
int EXT_API sfa_user_get_balance(struct sfa_user_info *sfa_user, int *balance, char *currency_buffer, size_t buffer_size);

const char * EXT_API sfa_buddy_state2str(enum sfa_buddy_state state) attr_const;
enum sfa_buddy_state EXT_API sfa_str2buddy_state(const char *str) attr_const;
enum sfa_buddy_state EXT_API sfa_user_get_buddy_state(struct sfa_user_info *sfa_user, const char *buddy, int buddy_autoadd);

enum sfa_buddy_state_update {
	SFA_BUDDY_STATE_UPDATE_NONE,
	SFA_BUDDY_STATE_UPDATE_ALL,
	SFA_BUDDY_STATE_UPDATE_SUBSCRIBED,
};

void EXT_API sfa_user_set_buddy_state_updates(struct sfa_user_info *sfa_user, enum sfa_buddy_state_update updates);

int EXT_API sfa_add_contact(struct sfa_user_info *sfa_user, const char *contactname);
int EXT_API sfa_remove_contact(struct sfa_user_info *sfa_user, const char *contactname);
int EXT_API sfa_request_contact_auth(struct sfa_user_info *sfa_user, const char *contactname, const char *authmsg);

enum sfa_account_status {
	SFA_ACCOUNT_STATUS_UNKNOWN,
	SFA_ACCOUNT_STATUS_LOGGED_OUT,
	SFA_ACCOUNT_STATUS_LOGGED_OUT_AND_PWD_SAVED,
	SFA_ACCOUNT_STATUS_CONNECTING_TO_P2P,
	SFA_ACCOUNT_STATUS_CONNECTING_TO_SERVER,
	SFA_ACCOUNT_STATUS_LOGGING_IN,
	SFA_ACCOUNT_STATUS_INITIALIZING,
	SFA_ACCOUNT_STATUS_LOGGED_IN,
	SFA_ACCOUNT_STATUS_LOGGING_OUT,
};


/*!
 * \brief Execute a function for every buddy in a Skype user's contact list
 * \param sfa_user User whose contact list is to be iterated over
 * \param buddy_func callback function to execute for each buddy
 * \param data data to pass to the callback function
 * \return nothing
 */
int EXT_API sfa_buddylist_iterate(struct sfa_user_info *sfa_user, void (*buddy_func)(struct client_user_info *, const char *, const char *, void *), void *data);
int EXT_API sfa_buddy_property_iterate(struct sfa_user_info *user, const char *buddy, void (*buddy_func)(struct client_user_info *, const char *, const char *, void *), void *data);

const char * EXT_API sfa_account_status2str(enum sfa_account_status status) attr_const;
const char * EXT_API sfa_account_logoutreason2str(enum sfa_account_logoutreason reason) attr_const;
int EXT_API sfa_user_connected(const struct sfa_user_info *sfa_user) attr_const;
int EXT_API sfa_user_stopped(const struct sfa_user_info *sfa_user) attr_const;
int EXT_API sfa_user_get_active_port(const struct sfa_user_info *sfa_user) attr_const;

int EXT_API sfa_user_loggedin(const struct sfa_user_info *sfa_user) attr_const;
int EXT_API sfa_user_loggedout(const struct sfa_user_info *sfa_user) attr_const;
const char * EXT_API sfa_user_account_status(const struct sfa_user_info *sfa_user) attr_const;
enum sfa_account_logoutreason EXT_API sfa_user_account_logoutreason(const struct sfa_user_info *sfa_user) attr_const;

void EXT_API sfa_user_set_debug(struct sfa_user_info *sfa_user, unsigned int value);
void EXT_API sfa_user_set_vedebug(struct sfa_user_info *sfa_user, unsigned int value);

enum sfa_auth_result {
	SFA_AUTH_ERROR = -1,
	SFA_AUTH_ACCEPTED,
	SFA_AUTH_DENIED,
	SFA_AUTH_BLOCKED,
	SFA_AUTH_IGNORED,
};

struct sfa_callbacks {
	/* called via sfa_user_run() when an incoming call request arrives */
	int (*new_call)(struct client_user_info *client_user, struct sfa_call_info *sfa_call, int control_socket, int chanalert);
	/* called via sfa_user_run() when a new chat message arrives */
	void (*new_chat_message)(const char *to, const char *from, const char *message);
	/* called via sfa_user_run() when an event has occurred on a call */
	void (*queue_event)(struct client_call_info *client_call, int frametype, int subclass, const char *reason);
	/* called via sfa_call_process() when control socket has been connected by voice engine */
	void (*control_socket_change)(struct client_call_info *client_call, int control_socket);
	/* called via sfa_user_run() when another Skype user is requesting presence authorization */
	enum sfa_auth_result (*authorization_request)(struct client_user_info *client_user, const char *buddy, const char *intro);
	/* called via sfa_call_process() when the voice engine has selected a codec for a call */
	void (*codec_selected)(struct client_call_info *client_call, const char *codec, unsigned int packet_size, unsigned int bit_rate);
	/* called via sfa_user_run() when a buddy availability change arrives */
	void (*buddy_state_change)(struct client_user_info *client_user, const char *buddy, enum sfa_buddy_state state);
	/* called via sfa_user_run() when the account status changes */
	void (*account_status_change)(struct client_user_info *client_user, enum sfa_account_status status);
	/* called via sfa_call_process() when the voice engine has selected an audio RTP address and port for a call */
	void (*audio_socket_address_selected)(struct client_call_info *client_call, const struct sockaddr_in *address);
	/* called via sfa_call_process() when the voice engine has determined the RFC2833 telephone-event payload type for a call */
	void (*dtmf_payload_type_selected)(struct client_call_info *client_call, uint8_t type);

	void (*logger)(int level, const char *file, int line, const char *function, const char *fmt, ...)
		       __attribute__((format(printf, 5, 6)));
	void (*verbose)(const char *fmt, ...)
		       __attribute__((format(printf, 1, 2)));
};

int EXT_API sfa_set_callbacks(struct sfa_callbacks *user_cb);
int EXT_API sfa_set_engine_directory(const char *path);
int EXT_API sfa_set_data_directory(const char *path);
int EXT_API sfa_set_license_directory(const char *path);
int EXT_API sfa_set_globals(unsigned int disable_tcpauto, unsigned int disable_udp,
	const struct sockaddr_in *local_addr, const struct sockaddr_in *rtp_addr,
	const char *hproxy, const char *hproxy_user, const char *hproxy_pass,
	const char *sproxy, const char *sproxy_user, const char *sproxy_pass);

enum sfa_startup_options {
	SFA_STARTUP_START_SKYHOST = (1 << 0),
};

int EXT_API sfa_startup(enum sfa_startup_options startup_options);
void EXT_API sfa_shutdown(void);
const char *sfa_get_version(void);

void EXT_API sfa_set_global_debug(unsigned int value);
unsigned int EXT_API sfa_get_online_count(void);

struct sfa_license;
struct sfa_license_iterator;

/*! \brief Create a new license iterator
 * \note This function locks the list of licenses, so always call sfa_license_iterator_destroy()
 *       both to free the memory and unlock the list of licenses
 * \return A newly allocated license iterator
 */
struct sfa_license_iterator * EXT_API sfa_license_iterator_new(void);

/*! \brief Destroy a license iterator created by sfa_license_iterator_new()
 * \param iter An iterator created by sfa_license_iterator_new()
 * \retval NULL
 */
struct sfa_license_iterator * EXT_API sfa_license_iterator_destroy(struct sfa_license_iterator *iter);

/*! \brief Retrieve the current license and advance the iterator
 * \param iter An iterator created by sfa_license_iterator_new()
 * \return The current license
 */
const struct sfa_license * EXT_API sfa_license_iterator_next(struct sfa_license_iterator *iter);

/*! \brief Retrieve a license's filename
 * \param license a license returned by sfa_license_iterator_next()
 * \return a C-style string containing the filename
 */
const char * EXT_API sfa_license_filename(const struct sfa_license *license);

/*! \brief Retrieve a license's key
 * \param license a license returned by sfa_license_iterator_next()
 * \return a C-style string containing the key
 */
const char * EXT_API sfa_license_key(const struct sfa_license *license);

/*! \brief Retrieve a license's hostid
 * \param license a license returned by sfa_license_iterator_next()
 * \return a C-style string containing the hostid
 */
const char * EXT_API sfa_license_hostid(const struct sfa_license *license);

/*! \brief Retrieve a license's expiration date
 * \param license a license returned by sfa_license_iterator_next()
 * \return a C-style string containing the expiration date
 */
const char * EXT_API sfa_license_expires(const struct sfa_license *license);

/*! \brief Retrieve a license's number of licensed channels
 * \param license a license returned by sfa_license_iterator_next()
 * \return the number of licensed channels
 */
int sfa_license_channels(const struct sfa_license *license);

/*! \brief Retrieve a license's status
 * \param license a license returned by sfa_license_iterator_next()
 * \returns a C-style string containing the status
 */
const char * EXT_API sfa_license_status(const struct sfa_license *license);

/*! \brief Return the calculated system Host-ID
 * \return a C-style string containing the calculated Host-ID
 */
const char * EXT_API sfa_get_hostid(void);

/*! \brief Return the total number of licensed SFA channels
 * \return total number of licensed SFA channels
 */
int EXT_API sfa_get_maxcalls(void);

/*! \brief Indicate that a module reload has been requested
 * \retval 0 Success
 */
int EXT_API sfa_on_reload(void);

/*! \brief Send a Skype chat message
 * \param sfa_user The skype account to senda message from
 * \param toname The skype account to send a message
 * \param message The message to send
 */
void EXT_API sfa_send_chat_message(struct sfa_user_info *user, const char *toname, const char *message);

/*! \brief Transfer a skype call to a another Skype user
 * \param call The skype call to transfer
 * \param toname Who to transfer to
 * \param topic An optional topic for the transfer
 * \retval 0 Success
 */
int EXT_API sfa_transfer_call(struct sfa_call_info *user, const char *toname, const char *topic);
#if defined(__cplusplus) || defined(c_plusplus)
}
#endif

#endif
